#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Common environment vars / settings
"""
import logging
import os
import re
import warnings
from pathlib import Path
from typing import List, Optional
from urllib.parse import urlparse

from pydantic import BaseSettings, Field, root_validator, validator

from argilla._constants import DEFAULT_MAX_KEYWORD_LENGTH, DEFAULT_TELEMETRY_KEY


class Settings(BaseSettings):
    """
    Main application settings. The pydantic BaseSettings class makes
    accessible environment variables by setting attributes.

    See <https://pydantic-docs.helpmanual.io/usage/settings/>

    only_bulk_api: (ONLY_BULK_API env var)
         If True, activate only bulk and search endpoints

    elasticseach: (ELASTICSEARCH env var)
        The elasticsearch endpoint for datasets persistence

    cors_origins: (CORS_ORIGINS env var)
        List of host patterns for CORS origin access

    docs_enabled: True
        If True, enable openapi docs endpoint at /api/docs

    es_records_index_shards:
        Configures the number of shards for dataset records index creation. Default=1

    es_records_index_replicas:
        Configures the number of shard replicas for dataset records index creation. Default=0

    disable_es_index_template_creation: (DISABLE_ES_INDEX_TEMPLATE_CREATION env var)
         Allowing advanced users to create their own es index settings and mappings. Default=False

    """

    __LOGGER__ = logging.getLogger(__name__)

    __DATASETS_INDEX_NAME__ = "ar.datasets"
    __DATASETS_RECORDS_INDEX_NAME__ = "ar.dataset.{}"

    home_path: Optional[str] = Field(description="The home path where argilla related files will be stored")
    base_url: Optional[str] = Field(description="The default base url where server will be deployed")
    database_url: Optional[str] = Field(description="The database url that argilla will use as data store")

    elasticsearch: str = "http://localhost:9200"
    elasticsearch_ssl_verify: bool = True
    elasticsearch_ca_path: Optional[str] = None
    cors_origins: List[str] = ["*"]

    docs_enabled: bool = True

    namespace: str = Field(default=None, regex=r"^[a-z]+$")

    enable_migration: bool = Field(
        default=False,
        description="If enabled, try to migrate data from old rubrix installation",
    )

    # Analyzer configuration
    default_es_search_analyzer: str = "standard"
    exact_es_search_analyzer: str = "whitespace"
    # This line will be enabled once words field won't be used anymore
    # wordcloud_es_search_analyzer: str = "multilingual_stop_analyzer"

    es_records_index_shards: int = 1
    es_records_index_replicas: int = 0

    vectors_fields_limit: int = Field(
        default=5,
        description="Max number of supported vectors per record",
    )

    metadata_fields_limit: int = Field(
        default=50,
        gt=0,
        le=100,
        description="Max number of fields in metadata",
    )
    metadata_field_length: int = Field(
        default=DEFAULT_MAX_KEYWORD_LENGTH,
        description="Max length supported for the string metadata fields."
        " Values containing higher than this will be truncated",
    )

    enable_telemetry: bool = True

    telemetry_key: str = DEFAULT_TELEMETRY_KEY

    @validator("home_path", always=True)
    def set_home_path_default(cls, home_path: str):
        return home_path or os.path.join(Path.home(), ".argilla")

    @validator("base_url", always=True)
    def normalize_base_url(cls, base_url: str):
        if not base_url:
            base_url = "/"
        if not base_url.startswith("/"):
            base_url = "/" + base_url
        if not base_url.endswith("/"):
            base_url += "/"

        return base_url

    @validator("database_url", pre=True, always=True)
    def set_database_url(cls, database_url: str, values: dict) -> str:
        if not database_url:
            home_path = values.get("home_path")
            sqlite_file = os.path.join(home_path, "argilla.db")
            return f"sqlite+aiosqlite:///{sqlite_file}?check_same_thread=False"

        if "sqlite" in database_url:
            regex = re.compile(r"sqlite(?!\+aiosqlite)")
            if regex.match(database_url):
                warnings.warn(
                    "From version 1.14.0, Argilla will use `aiosqlite` as default SQLite driver. The protocol in the"
                    " provided database URL has been automatically replaced from `sqlite` to `sqlite+aiosqlite`."
                    " Please, update your database URL to use `sqlite+aiosqlite` protocol."
                )
                return re.sub(regex, "sqlite+aiosqlite", database_url)

        if "postgresql" in database_url:
            regex = re.compile(r"postgresql(?!\+asyncpg)(\+psycopg2)?")
            if regex.match(database_url):
                warnings.warn(
                    "From version 1.14.0, Argilla will use `asyncpg` as default PostgreSQL driver. The protocol in the"
                    " provided database URL has been automatically replaced from `postgresql` to `postgresql+asyncpg`."
                    " Please, update your database URL to use `postgresql+asyncpg` protocol."
                )
                return re.sub(regex, "postgresql+asyncpg", database_url)

        return database_url

    @root_validator(skip_on_failure=True)
    def create_home_path(cls, values):
        Path(values["home_path"]).mkdir(parents=True, exist_ok=True)

        return values

    @property
    def dataset_index_name(self) -> str:
        ns = self.namespace
        if ns:
            return f"{self.namespace}.{self.__DATASETS_INDEX_NAME__}"
        return self.__DATASETS_INDEX_NAME__

    @property
    def dataset_records_index_name(self) -> str:
        ns = self.namespace
        if ns:
            return f"{self.namespace}.{self.__DATASETS_RECORDS_INDEX_NAME__}"
        return self.__DATASETS_RECORDS_INDEX_NAME__

    @property
    def old_dataset_index_name(self) -> str:
        index_name = ".rubrix<NAMESPACE>.datasets-v0"
        ns = self.namespace
        if ns is None:
            return index_name.replace("<NAMESPACE>", "")
        return index_name.replace("<NAMESPACE>", f".{ns}")

    @property
    def old_dataset_records_index_name(self) -> str:
        index_name = ".rubrix<NAMESPACE>.dataset.{}.records-v0"
        ns = self.namespace
        if ns is None:
            return index_name.replace("<NAMESPACE>", "")
        return index_name.replace("<NAMESPACE>", f".{ns}")

    def obfuscated_elasticsearch(self) -> str:
        """Returns configured elasticsearch url obfuscating the provided password, if any"""
        parsed = urlparse(self.elasticsearch)
        if parsed.password:
            return self.elasticsearch.replace(parsed.password, "XXXX")
        return self.elasticsearch

    class Config:
        env_prefix = "ARGILLA_"


settings = Settings()



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from fastapi import APIRouter, Depends, Query, Security

from argilla.server.apis.v0.handlers import metrics, token_classification_dataset_settings
from argilla.server.apis.v0.models.commons.model import BulkResponse
from argilla.server.apis.v0.models.commons.params import CommonTaskHandlerDependencies, RequestPagination
from argilla.server.apis.v0.models.token_classification import (
    TokenClassificationAggregations,
    TokenClassificationBulkRequest,
    TokenClassificationQuery,
    TokenClassificationRecord,
    TokenClassificationSearchRequest,
    TokenClassificationSearchResults,
)
from argilla.server.apis.v0.validators.token_classification import DatasetValidator
from argilla.server.commons.config import TasksFactory
from argilla.server.commons.models import TaskType
from argilla.server.errors import EntityNotFoundError
from argilla.server.models import User
from argilla.server.schemas.v0.datasets import CreateDatasetRequest
from argilla.server.security import auth
from argilla.server.services.datasets import DatasetsService
from argilla.server.services.tasks.token_classification import TokenClassificationService
from argilla.server.services.tasks.token_classification.metrics import TokenClassificationMetrics
from argilla.server.services.tasks.token_classification.model import (
    ServiceTokenClassificationQuery,
    ServiceTokenClassificationRecord,
)


def configure_router():
    task_type = TaskType.token_classification
    base_endpoint = f"/{{name}}/{task_type}"

    TasksFactory.register_task(
        task_type=task_type,
        query_request=TokenClassificationQuery,
        record_class=ServiceTokenClassificationRecord,
        metrics=TokenClassificationMetrics,
    )

    router = APIRouter(tags=[task_type], prefix="/datasets")

    @router.post(
        path=f"{base_endpoint}:bulk",
        operation_id="bulk_records",
        response_model=BulkResponse,
        response_model_exclude_none=True,
    )
    async def bulk_records(
        name: str,
        bulk: TokenClassificationBulkRequest,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TokenClassificationService = Depends(TokenClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        validator: DatasetValidator = Depends(DatasetValidator.get_instance),
        current_user: User = Security(auth.get_current_user, scopes=[]),
    ) -> BulkResponse:
        task = task_type
        workspace = common_params.workspace

        dataset = await datasets.find_by_name(current_user, name=name, task=task, workspace=workspace)

        await datasets.update(user=current_user, dataset=dataset, tags=bulk.tags, metadata=bulk.metadata)

        records = [ServiceTokenClassificationRecord.parse_obj(r) for r in bulk.records]
        # TODO(@frascuchon): validator can be applied in service layer
        await validator.validate_dataset_records(user=current_user, dataset=dataset, records=records)

        result = await service.add_records(dataset=dataset, records=records)
        return BulkResponse(dataset=name, processed=result.processed, failed=result.failed)

    @router.post(
        path=f"{base_endpoint}:search",
        response_model=TokenClassificationSearchResults,
        response_model_exclude_none=True,
        operation_id="search_records",
    )
    async def search_records(
        name: str,
        search: TokenClassificationSearchRequest = None,
        common_params: CommonTaskHandlerDependencies = Depends(),
        include_metrics: bool = Query(
            default=False,
            description="If enabled, return related record metrics",
        ),
        pagination: RequestPagination = Depends(),
        service: TokenClassificationService = Depends(TokenClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user, scopes=[]),
    ) -> TokenClassificationSearchResults:
        search = search or TokenClassificationSearchRequest()
        query = search.query or TokenClassificationQuery()

        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
        )
        results = service.search(
            dataset=dataset,
            query=ServiceTokenClassificationQuery.parse_obj(query),
            sort_by=search.sort,
            record_from=pagination.from_,
            size=pagination.limit,
            exclude_metrics=not include_metrics,
        )

        return TokenClassificationSearchResults(
            total=results.total,
            records=[TokenClassificationRecord.parse_obj(r) for r in results.records],
            aggregations=TokenClassificationAggregations.parse_obj(results.metrics) if results.metrics else None,
        )

    token_classification_dataset_settings.configure_router(router)
    metrics.configure_router(
        router,
        cfg=TasksFactory.get_task_by_task_type(task_type),
    )

    return router


router = configure_router()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import warnings
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Extra, Field, conint, conlist, root_validator, validator

from argilla.client.feedback.schemas.utils import LabelMappingMixin
from argilla.client.feedback.schemas.validators import title_must_have_value

QuestionTypes = Literal["text", "rating", "label_selection", "multi_label_selection", "ranking"]


class QuestionSchema(BaseModel):
    """Base schema for the `FeedbackDataset` questions. Which means that all the questions
    in the dataset will have at least these fields.

    Args:
        id: The ID of the question in Argilla. Defaults to None, and is automatically
            fulfilled internally once the question is pushed to Argilla.
        name: The name of the question. This is the only required field.
        title: The title of the question. If not provided, it will be capitalized from
            the `name` field. And its what will be shown in the UI.
        description: The description of the question. Defaults to None, and is not shown
            in the UI, otherwise, it will be shown in the tooltip close to each question.
        required: Whether the question is required or not. Defaults to True. Note that at
            least one question must be required.
        type: The type of the question. Defaults to None, and ideally it should be defined
            in the class inheriting from this one to be able to use a discriminated union
            based on the `type` field.
        settings: The settings of the question. Defaults to an empty dict, and it is
            automatically fulfilled internally before the question is pushed to Argilla,
            as the `settings` is part of the payload that will be sent to Argilla.

    Disclaimer:
        You should not use this class directly, but instead use the classes that inherit
        from this one, as they will have the `type` field already defined, and ensured
        to be supported by Argilla.
    """

    id: Optional[UUID] = None
    name: str = Field(..., regex=r"^(?=.*[a-z0-9])[a-z0-9_-]+$")
    title: Optional[str] = None
    description: Optional[str] = None
    required: bool = True
    type: Optional[QuestionTypes] = None
    settings: Dict[str, Any] = Field(default_factory=dict, allow_mutation=False)

    _title_must_have_value = validator("title", always=True, allow_reuse=True)(title_must_have_value)

    class Config:
        validate_assignment = True
        extra = Extra.forbid
        exclude = {"id", "type"}


# TODO(alvarobartt): add `XResponse` (e.g. `TextResponse`) classes
class TextQuestion(QuestionSchema):
    """Schema for the `FeedbackDataset` text questions, which are the ones that will
    require a text response from the user.

    Args:
        type: The type of the question. Defaults to 'text' and cannot/shouldn't be
            modified.
        use_markdown: Whether the question should be rendered using markdown or not.
            Defaults to False.

    Examples:
        >>> from argilla.client.feedback.schemas.questions import TextQuestion
        >>> TextQuestion(name="text_question", title="Text Question")
    """

    type: Literal["text"] = Field("text", allow_mutation=False)
    use_markdown: bool = False

    @root_validator(skip_on_failure=True)
    def update_settings(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        values["settings"]["type"] = values.get("type")
        values["settings"]["use_markdown"] = values.get("use_markdown", False)
        return values


class RatingQuestion(QuestionSchema, LabelMappingMixin):
    """Schema for the `FeedbackDataset` rating questions, which are the ones that will
    require a rating response from the user.

    Args:
        type: The type of the question. Defaults to 'rating' and cannot/shouldn't be
            modified.
        values: The list of interger values of the rating question. There is not need
            for the values to be sequential, but they must be unique, contain at least two
            unique integers in the range [1, 10].

    Examples:
        >>> from argilla.client.feedback.schemas.questions import RatingQuestion
        >>> RatingQuestion(name="rating_question", title="Rating Question", values=[1, 2, 3, 4, 5])
    """

    type: Literal["rating"] = Field("rating", allow_mutation=False)
    values: List[int] = Field(..., unique_items=True, ge=1, le=10, min_items=2)

    @root_validator(skip_on_failure=True)
    def update_settings(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        values["settings"]["type"] = values.get("type")
        values["settings"]["options"] = [{"value": value} for value in values.get("values")]
        return values


class _LabelQuestion(QuestionSchema, LabelMappingMixin):
    """Protected schema for the `FeedbackDataset` label questions, which are the ones that
    will require a label response from the user. This class should not be used directly,
    but instead use the classes that inherit from this one, which in this case are:
    `LabelSelectionQuestion` and `MultiLabelSelectionQuestion`.

    Args:
        type: The type of the question. Defaults to None and cannot/shouldn't be modified.
        labels: The list of labels of the label question. The labels must be unique, and
            the list must contain at least two unique labels. Additionally, `labels` can
            also be a dictionary of labels, where the keys are the labels, and the values
            are the labels that will be shown in the UI.
        visible_labels: The number of visible labels in the UI. Defaults to 20, and must
            be 3 or greater.
    """

    labels: Union[conlist(str, unique_items=True, min_items=2), Dict[str, str]]
    visible_labels: Optional[conint(ge=3)] = 20

    @validator("labels", always=True)
    def labels_dict_must_be_valid(cls, v: Union[List[str], Dict[str, str]]) -> Union[List[str], Dict[str, str]]:
        if isinstance(v, dict):
            assert len(v.keys()) > 1, "ensure this dict has at least 2 items"
            assert len(set(v.values())) == len(v.values()), "ensure this dict has unique values"
        return v

    @root_validator(skip_on_failure=True)
    def update_settings(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        values["settings"]["type"] = values.get("type")
        if isinstance(values.get("labels"), dict):
            values["settings"]["options"] = [
                {"value": key, "text": value} for key, value in values.get("labels").items()
            ]
        if isinstance(values.get("labels"), list):
            values["settings"]["options"] = [{"value": label, "text": label} for label in values.get("labels")]
        values["settings"]["visible_options"] = values.get(
            "visible_labels"
        )  # `None` is a possible value, which means all labels are visible
        return values


class LabelQuestion(_LabelQuestion):
    """Schema for the `FeedbackDataset` label questions, which are the ones that will
    require a label response from the user. This class should be used when the user can
    only select one label.

    Args:
        type: The type of the question. Defaults to 'label_selection' and cannot/shouldn't
            be modified.
        labels: The list of labels of the label question. The labels must be unique, and
            the list must contain at least two unique labels. Additionally, `labels` can
            also be a dictionary of labels, where the keys are the labels, and the values
            are the labels that will be shown in the UI.
        visible_labels: The number of visible labels in the UI. Defaults to 20, and must
            be 3 or greater.

    Examples:
        >>> from argilla.client.feedback.schemas.questions import LabelQuestion
        >>> LabelQuestion(name="label_question", title="Label Question", labels=["label_1", "label_2"])
    """

    type: Literal["label_selection"] = Field("label_selection", allow_mutation=False)


class MultiLabelQuestion(_LabelQuestion):
    """Schema for the `FeedbackDataset` label questions, which are the ones that will
    require a label response from the user. This class should be used when the user can
    select multiple labels.

    Args:
        type: The type of the question. Defaults to 'multi_label_selection' and
            cannot/shouldn't be modified.
        labels: The list of labels of the label question. The labels must be unique, and
            the list must contain at least two unique labels. Additionally, `labels` can
            also be a dictionary of labels, where the keys are the labels, and the values
            are the labels that will be shown in the UI.
        visible_labels: The number of visible labels in the UI. Defaults to 20, and must
            be 3 or greater.

    Examples:
        >>> from argilla.client.feedback.schemas.questions import MultiLabelQuestion
        >>> MultiLabelQuestion(name="multi_label_question", title="Multi Label Question", labels=["label_1", "label_2"])
    """

    type: Literal["multi_label_selection"] = Field("multi_label_selection", allow_mutation=False)


class RankingQuestion(QuestionSchema, LabelMappingMixin):
    """Schema for the `FeedbackDataset` ranking questions, which are the ones that will
    require a ranking response from the user. More specifically, the user will be asked
    to rank the labels, all the labels need to be assigned (if either the question is
    required or if at least one label has been ranked), and there can be ties/draws.

    Args:
        type: The type of the question. Defaults to 'ranking' and cannot/shouldn't be
            modified.
        values: The list of labels of the ranking question. The labels must be unique, and
            the list must contain at least two unique labels. Additionally, `labels` can
            also be a dictionary of labels, where the keys are the labels, and the values
            are the labels that will be shown in the UI.

    Examples:
        >>> from argilla.client.feedback.schemas.questions import RankingQuestion
        >>> RankingQuestion(name="ranking_question", title="Ranking Question", labels=["label_1", "label_2"])
    """

    type: Literal["ranking"] = Field("ranking", allow_mutation=False)
    values: Union[conlist(str, unique_items=True, min_items=2), Dict[str, str]]

    @validator("values", always=True)
    def values_dict_must_be_valid(cls, v: Union[List[str], Dict[str, str]]) -> Union[List[str], Dict[str, str]]:
        if isinstance(v, dict):
            assert len(v.keys()) > 1, "ensure this dict has at least 2 items"
            assert len(set(v.values())) == len(v.values()), "ensure this dict has unique values"
        return v

    @root_validator(skip_on_failure=True)
    def update_settings(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        values["settings"]["type"] = values.get("type")
        if isinstance(values.get("values"), dict):
            values["settings"]["options"] = [
                {"value": key, "text": value} for key, value in values.get("values").items()
            ]
        if isinstance(values.get("values"), list):
            values["settings"]["options"] = [{"value": value, "text": value} for value in values.get("values")]
        return values



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""create suggestions table

Revision ID: 3fc3c0839959
Revises: 8c574ada5e5f
Create Date: 2023-06-27 17:34:49.734260

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3fc3c0839959"
down_revision = "8c574ada5e5f"
branch_labels = None
depends_on = None

suggestion_type_enum = sa.Enum("model", "human", name="suggestion_type_enum")


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "suggestions",
        sa.Column("id", sa.Uuid(), nullable=False),
        sa.Column("value", sa.JSON(), nullable=False),
        sa.Column("score", sa.Float(), nullable=True),
        sa.Column("agent", sa.String(), nullable=True),
        sa.Column("type", sa.Enum(name="suggestion_type_enum"), nullable=True),
        sa.Column("record_id", sa.Uuid(), nullable=False),
        sa.Column("question_id", sa.Uuid(), nullable=False),
        sa.Column("inserted_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["question_id"], ["questions.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["record_id"], ["records.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("record_id", "question_id", name="suggestion_record_id_question_id_uq"),
    )
    op.create_index(op.f("ix_suggestions_question_id"), "suggestions", ["question_id"], unique=False)
    op.create_index(op.f("ix_suggestions_record_id"), "suggestions", ["record_id"], unique=False)
    op.create_index(op.f("ix_suggestions_type"), "suggestions", ["type"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    bind = op.get_bind()

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_suggestions_type"), table_name="suggestions")
    op.drop_index(op.f("ix_suggestions_record_id"), table_name="suggestions")
    op.drop_index(op.f("ix_suggestions_question_id"), table_name="suggestions")
    op.drop_table("suggestions")
    # ### end Alembic commands ###

    # For this see here: https://github.com/sqlalchemy/alembic/issues/886
    if bind.dialect.name == "postgresql":
        suggestion_type_enum.drop(bind)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

from prodict import Prodict

from argilla.client.models import Record

if TYPE_CHECKING:
    from argilla.listeners import RGDatasetListener


@dataclasses.dataclass
class Search:
    """
    Search results for a single listener execution

    Args:
        total: The total number of records affected by the listener query
        query_params: The query parameters applied to the executed search
    """

    total: int
    query_params: Optional[Dict[str, Any]] = None


class Metrics(Prodict):
    """
    Metrics results for a single listener execution.

    The metrics object exposes the metrics configured for the listener as property values.
    For example, if you define a listener including the metric "F1", the results will be
    accessible as ``metrics.F1``
    """

    pass


@dataclasses.dataclass
class RGListenerContext:
    """
    The argilla listener execution context. This class keeps the context components related to a listener

    Args:

        listener: The argilla listener instance
        search: Search results for current execution
        metrics: Metrics results for current execution
        query_params: Dynamic parameters used in the listener query
    """

    listener: "RGDatasetListener" = dataclasses.field(repr=False, hash=False)
    search: Optional[Search] = None
    metrics: Optional[Metrics] = None
    query_params: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        self.__listener__ = self.listener
        del self.listener

    @property
    def dataset(self) -> str:
        """Computed property that returns the configured listener dataset name"""
        return self.__listener__.dataset

    @property
    def query(self) -> Optional[str]:
        """Computed property that returns the configured listener query string"""
        return self.__listener__.formatted_query


ListenerCondition = Callable[[Search, Optional[RGListenerContext]], bool]
ListenerAction = Union[
    Callable[[List[Record], RGListenerContext], bool],
    Callable[[RGListenerContext], bool],
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

from opensearchpy import OpenSearch, helpers
from opensearchpy.exceptions import NotFoundError, OpenSearchException, OpenSearchWarning, RequestError
from opensearchpy.helpers import BulkIndexError

from argilla.server.daos.backend import query_helpers
from argilla.server.daos.backend.base import BackendErrorHandler, IndexNotFoundError
from argilla.server.daos.backend.client_adapters.base import IClientAdapter
from argilla.server.daos.backend.metrics.base import ElasticsearchMetric
from argilla.server.daos.backend.search.model import BaseQuery, SortableField, SortConfig
from argilla.server.daos.backend.search.query_builder import HighlightParser, OpenSearchQueryBuilder


@dataclasses.dataclass
class OpenSearchClient(IClientAdapter):
    index_shards: int

    config_backend: Dict[str, Any]
    highlight = HighlightParser()

    query_builder = OpenSearchQueryBuilder()

    def __post_init__(self):
        self.__client__ = OpenSearch(**self.config_backend)
        self.error_handling = BackendErrorHandler(
            WarningIgnore=OpenSearchWarning,
            RequestError=RequestError,
            BulkError=BulkIndexError,
            NotFoundError=NotFoundError,
            GenericApiError=OpenSearchException,
        )

    def configure_index_vectors(
        self,
        *,
        index: str,
        vectors: Dict[str, int],
    ):
        self._check_vector_supported()

        self.set_index_settings(
            index=index,
            settings={"index.knn": False},
        )
        vector_mappings = {}
        for vector_name, vector_dimension in vectors.items():
            index_mapping = {
                "type": "knn_vector",
                "dimension": vector_dimension,
                "method": {
                    "name": "hnsw",
                    "engine": "lucene",
                    "space_type": "l2",
                    "parameters": {"m": 2, "ef_construction": 4},
                },
            }
            vector_field = self.query_builder.get_vector_field_name(vector_name)
            vector_mappings[vector_field] = index_mapping

        self.set_index_mappings(
            index=index,
            properties=vector_mappings,
        )

    def _check_vector_supported(self):
        if not self.vector_search_supported:
            raise ValueError(
                "The vector search is not supported for this elasticsearch version. "
                "Please, update the server to use this feature"
            )

    def search_docs(
        self,
        *,
        index: str,
        query: Optional[BaseQuery] = None,
        sort: Optional[SortConfig] = None,
        doc_from: int = 0,
        size: int = 100,
        exclude_fields: List[str] = None,
        enable_highlight: bool = True,
        routing: str = None,
    ) -> Dict[str, Any]:
        with self.error_handling(index=index):
            highlight = self.highlight if enable_highlight else None
            es_query = self.query_builder.map_2_es_query(
                schema=self.get_index_schema(index=index),
                query=query,
                sort=sort,
                exclude_fields=exclude_fields,
                doc_from=doc_from,
                highlight=highlight,
                size=size,
            )

            results = self._es_search(
                index=index,
                es_query=es_query,
                size=size,
                routing=routing,
            )

            return self._process_search_results(
                search_results=results,
                highlight_parser=highlight,
            )

    def bulk(
        self,
        *,
        index: str,
        actions: Iterable[dict],
    ) -> Tuple:
        with self.error_handling(index=index):
            return helpers.bulk(
                client=self.__client__,
                index=index,
                actions=actions,
                raise_on_error=True,
                refresh="wait_for",
            )

    def _reindex(
        self,
        *,
        source_index: str,
        target_index: str,
    ):
        with self.error_handling(index=source_index):
            helpers.reindex(
                client=self.__client__,
                source_index=source_index,
                target_index=target_index,
            )

    def get_cluster_info(self) -> Dict[str, Any]:
        """Returns basic about es cluster"""
        with self.error_handling():
            return self.__client__.info()

    def get_index_schema(self, *, index: str):
        with self.error_handling(index=index):
            response = self.__client__.indices.get_mapping(index=index)
            if index in response:
                return response.get(index)
            elif len(response) == 1:
                return list(response.values())[0]
            return response

    def drop_document_property(
        self,
        index: str,
        id: str,
        property: str,
    ):
        self.upsert_index_document(
            index=index,
            id=id,
            script=f'ctx._source.remove("{property}")',
            partial_update=True,
        )

    def update_docs_by_query(
        self,
        *,
        index: str,
        data: Dict[str, Any],
        query: Optional[BaseQuery] = None,
    ) -> Dict[str, Any]:
        es_query = self.query_builder.map_2_es_query(
            schema=self.get_index_schema(index=index),
            query=query,
        )
        return self._update_by_query(
            index=index,
            query=es_query["query"],
            data=data,
        )

    def compute_index_metric(
        self,
        *,
        index: str,
        metric: ElasticsearchMetric,
        query: BaseQuery,
        params: Dict[str, Any],
    ):
        params.update(
            {
                "index": index,
                "client": self,
            }
        )

        filtered_params = {argument: params[argument] for argument in metric.metric_arg_names if argument in params}

        aggregations = metric.aggregation_request(**filtered_params)
        if not aggregations:
            return {}
        if not isinstance(aggregations, list):
            aggregations = [aggregations]

        with self.error_handling(index=index):
            es_query = self.query_builder.map_2_es_query(
                schema=self.get_index_schema(index=index),
                query=query,
            )

            results = {}
            for aggregation in aggregations:
                es_query["aggs"] = aggregation

                search = self._es_search(
                    index=index,
                    es_query=es_query,
                    size=0,
                )

                search_aggregations = search.get("aggregations", {})
                if search_aggregations:
                    parsed_aggregations = query_helpers.parse_aggregations(search_aggregations)
                    results.update(parsed_aggregations)
            return metric.aggregation_result(results.get(metric.id, results))

    def delete_docs_by_query(
        self,
        *,
        index: str,
        query: Optional[BaseQuery],
    ) -> Tuple[int, int]:
        es_query = self.query_builder.map_2_es_query(
            schema=self.get_index_schema(index=index),
            query=query,
        )
        return self._delete_by_query(
            index=index,
            query=es_query["query"],
        )

    def _delete_by_query(
        self,
        *,
        index: str,
        query: Dict[str, Any],
    ) -> Tuple[int, int]:
        with self.error_handling(index=index):
            response = self.__client__.delete_by_query(
                index=index,
                body={"query": query},
                slices="auto",
                wait_for_completion=True,
                # conflicts="proceed",  # If document version conflict -> continue
            )
            total, deleted = response["total"], response["deleted"]
            return total, deleted

    def _update_by_query(
        self,
        *,
        index,
        query: Dict[str, Any],
        data: Dict[str, Any],
    ) -> Dict[str, Any]:
        with self.error_handling(index=index):
            inline = ";".join([f"ctx._source.{k}='{v}'" for k, v in data.items()])
            response = self.__client__.update_by_query(
                index=index,
                body={
                    "query": query,
                    "script": {
                        "lang": "painless",
                        "inline": inline,
                    },
                },
                slices="auto",
                wait_for_completion=True,
                conflicts="proceed",
            )
            return {
                "total": response["total"],
                "updated": response["updated"],
            }

    def scan_docs(
        self,
        index: str,
        query: BaseQuery,
        sort: SortConfig,
        size: Optional[int] = None,
        fetch_once: bool = False,
        search_from_params: Optional[Any] = None,
        enable_highlight: bool = False,
        include_fields: Optional[List[str]] = None,
        exclude_fields: Optional[List[str]] = None,
    ) -> Iterable[Dict[str, Any]]:
        batch_size = size or 500

        highlight = self.highlight if enable_highlight else None
        es_query = self.query_builder.map_2_es_query(
            query=query,
            schema=self.get_index_schema(index=index),
            sort=sort,
            search_after_param=search_from_params,
            include_fields=include_fields,
            exclude_fields=exclude_fields,
            highlight=highlight,
        )
        es_query = es_query.copy() or {}
        response = self.__client__.search(
            index=index,
            body=es_query,
            size=batch_size,
            track_total_hits=False,
        )
        records_yield = 0
        while response["hits"]["hits"]:
            hit = None
            for hit in response["hits"]["hits"]:
                yield self._normalize_document(document=hit, highlight=highlight, add_sort_info=True)
                records_yield += 1

            if fetch_once or (size and size >= records_yield):
                break

            next_search_from = hit["sort"]
            es_query["search_after"] = next_search_from
            response = self.__client__.search(index=index, body=es_query, size=size, track_total_hits=False)

    def _process_search_results(
        self,
        *,
        search_results: Dict[str, Any],
        highlight_parser: Optional[HighlightParser] = None,
    ):
        hits = search_results["hits"]
        total = hits["total"]
        docs = hits["hits"]
        return {
            "total": total,
            "docs": [
                self._normalize_document(
                    document=doc,
                    highlight=highlight_parser,
                )
                for doc in docs
            ],
        }

    def copy_index(
        self,
        *,
        source_index: str,
        target_index: str,
        override: bool = True,
        reindex: bool = False,
    ):
        source_index = self._get_original_index_name(source_index)

        if reindex:
            return self._reindex(source_index=source_index, target_index=target_index)

        is_source_read_only_index = self._is_read_only_index(index=source_index)

        try:
            if not is_source_read_only_index:
                self._enable_or_disable_read_only_index(index=source_index, read_only=True)

            if override:
                self.delete_index(index=target_index)

            self._clone_index(index_from=source_index, index_to=target_index)
        finally:
            self._enable_or_disable_read_only_index(index=source_index, read_only=is_source_read_only_index)
            self._enable_or_disable_read_only_index(index=target_index, read_only=is_source_read_only_index)

    def _clone_index(
        self,
        *,
        index_from: str,
        index_to: str,
    ):
        with self.error_handling(index=index_from):
            self.__client__.indices.clone(
                index=index_from,
                target=index_to,
                wait_for_active_shards=self.index_shards,
            )

    def get_property_type(
        self,
        *,
        index: str,
        property_name: str,
        drop_extra_props: bool = False,
    ) -> Dict[str, str]:
        try:
            schema = self._get_fields_schema(
                index=index,
                fields=f"{property_name}.*",
            )

            if drop_extra_props:
                # Remove `text`, `exact` and `wordcloud` fields
                def is_subfield(key: str):
                    for suffix in ["exact", "text", "wordcloud"]:
                        if suffix in key:
                            return True
                    return False

                schema = {key: value for key, value in schema.items() if not is_subfield(key)}

            return schema
        except IndexNotFoundError:
            # No mapping data
            return {}

    def get_index_document_by_id(
        self,
        index: str,
        id: str,
    ) -> Optional[Dict[str, Any]]:
        try:
            with self.error_handling(index=index):
                if self.__client__.exists(
                    index=index,
                    id=id,
                ):
                    doc = self.__client__.get(index=index, id=id)
                    return self._normalize_document(document=doc)
        # TODO: Review this exception when not found an document instead of an index
        except IndexNotFoundError:
            return None

    def create_index_alias(
        self,
        *,
        index: str,
        alias: str,
    ):
        with self.error_handling(index=index):
            self.__client__.indices.put_alias(
                index=index,
                name=alias,
                ignore=[400, 404],
            )

    def delete_index_document(
        self,
        index: str,
        id: str,
    ):
        with self.error_handling(index=index):
            if self.__client__.exists(
                index=index,
                id=id,
            ):
                self.__client__.delete(
                    index=index,
                    id=id,
                    refresh=True,
                )

    def delete_index_alias(
        self,
        *,
        index: str,
        alias: str,
    ):
        with self.error_handling(index=index):
            self.__client__.indices.delete_alias(
                index=index,
                name=alias,
                ignore=[400, 404],
            )

    def exists_index(self, index: str) -> bool:
        with self.error_handling(index=index):
            return self.__client__.indices.exists(index=index)

    def delete_index(
        self,
        *,
        index: str,
        raises_error: bool = False,
    ):
        with self.error_handling(index=index):
            if self.exists_index(index):
                ignore_errors = [400, 404]
                if raises_error:
                    ignore_errors.clear()
                self.__client__.indices.delete(
                    index=index,
                    ignore=ignore_errors,
                )

    def create_index(
        self,
        index: str,
        force_recreate: bool = False,
        settings: Dict[str, Any] = None,
        mappings: Dict[str, Any] = None,
    ):
        with self.error_handling(index):
            if force_recreate:
                self.delete_index(index=index)
            if not self.exists_index(index=index):
                self.__client__.indices.create(
                    index=index,
                    body={
                        "settings": settings or {},
                        "mappings": mappings or {},
                    },
                    ignore=400,
                )

    def index_documents(self, index: str, docs: List[Dict[str, Any]]) -> int:
        actions = (self._doc2bulk_action(index, doc) for doc in docs)
        success, failed = self.bulk(
            index=index,
            actions=actions,
        )
        return len(failed)

    @staticmethod
    def _doc2bulk_action(index: str, doc: Dict[str, Any]) -> Dict[str, Any]:
        doc_id = doc.get("id")

        data = (
            {"_index": index, "_op_type": "index", **doc}
            if doc_id is None
            else {"_index": index, "_id": doc_id, "_op_type": "update", "doc_as_upsert": True, "doc": doc}
        )

        return data

    def upsert_index_document(
        self,
        index: str,
        id: str,
        document: Optional[Dict[str, Any]] = None,
        script: Optional[str] = None,
        partial_update: bool = False,
    ):
        # TODO: validate either doc or script are provided
        with self.error_handling(index=index):
            if not partial_update:
                return self.index_document(
                    index=index,
                    id=id,
                    body=document,
                )

            return self.update_document(
                index=index,
                id=id,
                body={"script": script} if script else {"doc": document},
            )

    def index_document(
        self,
        *,
        index: str,
        id: str,
        body: dict,
    ):
        with self.error_handling(index=index):
            return self.__client__.index(
                index=index,
                id=id,
                body=body,
                refresh=True,
            )

    def update_document(
        self,
        *,
        index: str,
        id: str,
        body: dict,
    ):
        with self.error_handling(index=index):
            return self.__client__.update(
                index=index,
                id=id,
                body=body,
                refresh=True,
                retry_on_conflict=500,  # TODO: configurable
            )

    def open_index(self, index: str):
        with self.error_handling(index=index):
            self.__client__.indices.open(
                index=index,
                wait_for_active_shards=self.index_shards,
            )

    def close_index(self, index: str):
        with self.error_handling(index=index):
            self.__client__.indices.close(
                index=index,
                ignore_unavailable=True,
                wait_for_active_shards=self.index_shards,
            )

    def clone_index(
        self,
        source_index: str,
        target_index: str,
    ):
        with self.error_handling(index=source_index):
            self.__client__.indices.clone(
                index=source_index,
                target=target_index,
                wait_for_active_shards=self.index_shards,
            )

    def _is_read_only_index(self, index: str) -> bool:
        with self.error_handling(index=index):
            response = self.__client__.indices.get_settings(
                index=index,
                name="index.blocks.write",
                allow_no_indices=True,
                flat_settings=True,
            )
            if not response:
                return False

            return response[index]["settings"]["index.blocks.write"] == "true"

    def set_index_settings(
        self,
        index: str,
        settings: Dict[str, Any],
    ):
        with self.error_handling(index=index):
            try:
                self.close_index(index)
                self.__client__.indices.put_settings(
                    index=index,
                    body=settings,
                )
            finally:
                self.open_index(index)

    def set_index_mappings(
        self,
        index: str,
        properties: Optional[Dict[str, Any]] = None,
        **extra_cfg,
    ):
        with self.error_handling(index=index):
            body = extra_cfg or {}
            if properties:
                body.update({"properties": properties})

            self.__client__.indices.put_mapping(
                index=index,
                body=body,
            )

    def _enable_or_disable_read_only_index(
        self,
        *,
        index: str,
        read_only: bool,
    ):
        with self.error_handling(index=index):
            self.__client__.indices.put_settings(
                index=index,
                body={"settings": {"index.blocks.write": read_only}},
                ignore=404,
            )

    def _get_fields_schema(
        self,
        *,
        index: str,
        fields: str,
    ):
        with self.error_handling(index=index):
            response = self.__client__.indices.get_field_mapping(
                fields=fields,
                index=index,
                ignore_unavailable=False,
            )
            if index in response:
                data = response.get(index)
            elif len(response) == 1:
                data = list(response.values())[0]
            else:
                data = response

            return {
                key: list(definition["mapping"].values())[0]["type"] for key, definition in data["mappings"].items()
            }

    def _normalize_document(
        self,
        document: Dict[str, Any],
        highlight: Optional[HighlightParser] = None,
        is_phrase_query: bool = True,
        add_sort_info: bool = False,
    ):
        data = document["_source"]

        if "id" not in data:
            data["id"] = document["_id"]

        if add_sort_info and "sort" in document:
            data["sort"] = document["sort"]

        if highlight:
            keywords = highlight.parse_highligth_results(
                doc=document,
                is_phrase_query=is_phrase_query,
            )
            if keywords:
                data[highlight.search_keywords_field] = keywords

        return data

    def _es_search(
        self,
        index: str,
        es_query: Dict[str, Any],
        size: int,
        routing: Optional[str] = None,
    ):
        if "knn" in es_query["query"]:
            self._check_vector_supported()

        results = self.__client__.search(
            index=index,
            body=es_query,
            routing=routing,
            track_total_hits=True,
            rest_total_hits_as_int=True,
            size=size,
        )
        return results

    def _get_original_index_name(self, source_index_or_alias: str):
        response = self.__client__.indices.get(index=source_index_or_alias)
        # The response contains a single key with the original index name (if we provided an alias)
        for index_name in response:
            return index_name



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import warnings
from collections import Counter
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from argilla import load
from argilla.client.datasets import DatasetForTextClassification
from argilla.client.models import TextClassificationRecord
from argilla.labeling.text_classification.rule import Rule, load_rules


def _add_docstr(string: str):
    def docstring_decorator(fn):
        fn.__doc__ = string
        return fn

    return docstring_decorator


class WeakLabelsBase:
    """Base class for the weak label classes.

    Tries to facilitate the implementations, avoids code duplications.

    Args:
        dataset: Name of the dataset to which the rules will be applied.
        rules: A list of rules (labeling functions). They must return a string, or ``None`` in case of abstention.
            If None, we will use the rules of the dataset (Default).
        ids: An optional list of record ids to filter the dataset before applying the rules.
        query: An optional ElasticSearch query with the
            `query string syntax <https://argilla.readthedocs.io/en/stable/guides/queries.html>`_
            to filter the dataset before applying the rules.

    Raises:
        NoRulesFoundError: When you do not provide rules, and the dataset has no rules either.
        DuplicatedRuleNameError: When you provided multiple rules with the same name.
        NoRecordsFoundError: When the filtered dataset is empty.
    """

    def __init__(
        self,
        dataset: str,
        rules: Optional[List[Callable]] = None,
        ids: Optional[List[Union[int, str]]] = None,
        query: Optional[str] = None,
    ):
        if not isinstance(dataset, str):
            raise TypeError(f"The name of the dataset must be a string, but you provided: {dataset}")
        self._dataset = dataset

        self._rules = rules or load_rules(dataset)
        if not self._rules:
            raise NoRulesFoundError(f"No rules were found in the given dataset '{dataset}'")

        self._rules_index2name = {
            # covers our Rule class, snorkel's LabelingFunction class and arbitrary methods
            index: (
                getattr(rule, "name", None)
                or (
                    getattr(rule, "__name__", None)
                    # allow multiple lambda functions
                    if getattr(rule, "__name__", None) != "<lambda>"
                    else None
                )
                or f"rule_{index}"
            )
            for index, rule in enumerate(self._rules)
        }
        # raise error if there are duplicates
        counts = Counter(self._rules_index2name.values())
        if len(counts.keys()) < len(self._rules):
            raise DuplicatedRuleNameError(
                f"Following rule names are duplicated x times: { {key: val for key, val in counts.items() if val > 1} }"
                " Please make sure to provide unique rule names."
            )
        self._rules_name2index = {val: key for key, val in self._rules_index2name.items()}

        # load records and check compatibility
        self._records: DatasetForTextClassification = load(dataset, query=query, ids=ids, include_vectors=False)
        if not self._records:
            raise NoRecordsFoundError(
                f"No records found in dataset '{dataset}'"
                + (f" with query '{query}'" if query else "")
                + (" and" if query and ids else "")
                + (f" with ids {ids}." if ids else ".")
            )

        self._matrix = self._extended_matrix = self._extension_queries = None

    @property
    def rules(self) -> List[Callable]:
        """The rules (labeling functions) that were used to produce the weak labels."""
        return self._rules

    @property
    def labels(self) -> List[str]:
        """The list of labels."""
        raise NotImplementedError

    @property
    def cardinality(self) -> int:
        """The number of labels."""
        raise NotImplementedError

    def records(self, has_annotation: Optional[bool] = None) -> List[TextClassificationRecord]:
        """Returns the records corresponding to the weak label matrix

        Args:
            has_annotation: If True, return only the records that have an annotation. If False, return only the records
                that have NO annotation. By default, we return all the records.

        Returns:
            A list of records, or optionally just a part of them.
        """
        if has_annotation is True:
            return [rec for rec in self._records if rec.annotation is not None]
        if has_annotation is False:
            return [rec for rec in self._records if rec.annotation is None]

        return list(self._records)

    def matrix(self, has_annotation: Optional[bool] = None) -> np.ndarray:
        """Returns the weak label matrix, or optionally just a part of it.

        Args:
            has_annotation: If True, return only the part of the matrix that has a corresponding annotation.
                If False, return only the part of the matrix that has NOT a corresponding annotation.
                By default, we return the whole weak label matrix.

        Returns:
            The weak label matrix, or optionally just a part of it.
        """
        raise NotImplementedError

    def annotation(
        self,
        include_missing: bool = False,
    ) -> np.ndarray:
        """Returns the annotation labels.

        Args:
            include_missing: If True, returns an array of the length of the record list (``self.records()``).

        Returns:
            The annotation labels.
        """
        raise NotImplementedError

    def summary(
        self,
        normalize_by_coverage: bool = False,
        annotation: Optional[np.ndarray] = None,
    ) -> pd.DataFrame:
        """Returns a summary statistics for each rule:

        Args:
            normalize_by_coverage: Normalize the overlaps and conflicts by the respective coverage.
            annotation: An optional array holding the annotations.
                By default, we will use ``self.annotation(include_missing=True)``.

        Returns:
            The summary statistics for each rule in a pandas DataFrame.
        """
        raise NotImplementedError

    def show_records(
        self,
        labels: Optional[List[str]] = None,
        rules: Optional[List[Union[str, int]]] = None,
    ) -> pd.DataFrame:
        """Shows records in a pandas DataFrame, optionally filtered by weak labels and non-abstaining rules.

        If you provide both ``labels`` and ``rules``, we take the intersection of both filters.

        Args:
            labels: All of these labels are in the record's weak labels. If None, do not filter by labels.
            rules: All of these rules did not abstain for the record. If None, do not filter by rules.
                You can refer to the rules by their (function) name or by their index in the ``self.rules`` list.

        Returns:
            The optionally filtered records as a pandas DataFrame.
        """
        raise NotImplementedError

    def _compute_overlaps_conflicts(
        self,
        has_weak_label: np.ndarray,
        has_overlaps_or_conflicts: np.ndarray,
        coverage: np.ndarray,
        normalize_by_coverage: bool,
    ) -> np.ndarray:
        """Helper method to compute the overlaps/conflicts and optionally normalize them by the respective coverage

        Args:
            has_weak_label: 2D boolean matrix (i, j) that indicates if the record i has a weak label from rule j.
            has_overlaps_or_conflicts: Array that indicates if the record has overlapping/conflicting weak labels.
            coverage: Array of coverages for each rule
            normalize_by_coverage: Normalize the overlaps/conflicts by the respective coverage.

        Returns:
            Array of fractions of overlaps/conflicts for each rule, optionally normalized by their coverages.
        """
        overlaps_or_conflicts = (
            has_weak_label * np.repeat(has_overlaps_or_conflicts, len(self._rules)).reshape(has_weak_label.shape)
        ).sum(axis=0) / len(self._records)
        # total
        overlaps_or_conflicts = np.append(overlaps_or_conflicts, has_overlaps_or_conflicts.sum() / len(self._records))

        if normalize_by_coverage:
            # ignore division by 0 warnings, as we convert the nan back to 0.0 afterwards
            with np.errstate(divide="ignore", invalid="ignore"):
                overlaps_or_conflicts /= coverage
            return np.nan_to_num(overlaps_or_conflicts)

        return overlaps_or_conflicts

    def extend_matrix(
        self,
        thresholds: Union[List[float], np.ndarray],
        embeddings: Optional[np.ndarray] = None,
        gpu: bool = False,
    ):
        """Extends the weak label matrix through embeddings according to the similarity thresholds for each rule.

        Implementation based on `Epoxy <https://github.com/HazyResearch/epoxy>`__.

        Args:
            thresholds: An array of thresholds between 0.0 and 1.0, one for each column of the weak labels matrix.
                Each one stands for the minimum cosine similarity between two sentences for a rule to be extended.
            embeddings: Embeddings for each row of the weak label matrix.
                If not provided, we will use the ones from the last ``WeakLabels.extend_matrix()`` call.
            gpu: If True, perform FAISS similarity queries on GPU.

        Examples:
            >>> # Choose any model to generate the embeddings.
            >>> from sentence_transformers import SentenceTransformer
            >>> model = SentenceTransformer('all-mpnet-base-v2', device='cuda')
            >>>
            >>> # Generate the embeddings and set the thresholds.
            >>> weak_labels = {class_name}(dataset="my_dataset")
            >>> embeddings = np.array([ model.encode(rec.text) for rec in weak_labels.records() ])
            >>> thresholds = [0.6] * len(weak_labels.rules)
            >>>
            >>> # Extend the weak labels matrix.
            >>> weak_labels.extend_matrix(thresholds, embeddings)
            >>>
            >>> # Calling the method below will now retrieve the extended matrix.
            >>> weak_labels.matrix()
            >>>
            >>> # Subsequent calls without the embeddings parameter will reutilize the faiss index built on the first call.
            >>> thresholds = [0.75] * len(weak_labels.rules)
            >>> weak_labels.extend_matrix(thresholds)
            >>> weak_labels.matrix()
        """
        abstains, supports = self._extend_matrix_preprocess()

        if embeddings is not None:
            self._extension_queries = self._find_dists_and_nearest(
                np.copy(embeddings).astype(np.float32), abstains, supports, gpu=gpu
            )
        elif self._extension_queries is None:
            raise ValueError("Embeddings are not optional the first time a matrix is extended.")
        dists, nearest = self._extension_queries

        self._extended_matrix = np.copy(self._matrix)
        new_points = [(dists[i] > thresholds[i]) for i in range(self._matrix.shape[1])]
        for i in range(self._matrix.shape[1]):
            self._extended_matrix[abstains[i][new_points[i]], i] = self._matrix[supports[i], i][
                nearest[i][new_points[i]]
            ]

        self._extend_matrix_postprocess()

    def _find_dists_and_nearest(
        self,
        embeddings: np.ndarray,
        abstains: List[np.ndarray],
        support: List[np.ndarray],
        gpu: bool,
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Helper method to extend the weak labels."""
        try:
            import faiss
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "'faiss' must be installed to extend a weak label matrix! "
                "You can install 'faiss' with the commands: `pip install faiss-cpu` or `pip install faiss-gpu`"
            )
        faiss.normalize_L2(embeddings)
        embeddings_length = embeddings.shape[1]

        label_fn_indexes = [faiss.IndexFlatIP(embeddings_length) for i in range(self._matrix.shape[1])]

        if gpu:
            res = faiss.StandardGpuResources()
            label_fn_indexes = [faiss.index_cpu_to_gpu(res, 0, x) for x in label_fn_indexes]

        for i in range(self._matrix.shape[1]):
            label_fn_indexes[i].add(embeddings[support[i]])

        dists_and_nearest = []
        for i in tqdm(range(self._matrix.shape[1]), total=self._matrix.shape[1]):
            embs_query = np.copy(embeddings[abstains[i]])
            faiss.normalize_L2(embs_query)
            dists_and_nearest.append(label_fn_indexes[i].search(embs_query, 1))

        dists = [dist_and_nearest[0].flatten() for dist_and_nearest in dists_and_nearest]
        nearest = [dist_and_nearest[1].flatten() for dist_and_nearest in dists_and_nearest]

        return dists, nearest

    def _extend_matrix_preprocess(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Helper method to compute the abstains and supports.

        abstains: List of record indices per rule for which the rule abstained.
        supports: List of record indices per rule for which the rule voted.

        Returns:
            A tuple containing the abstains and supports
        """
        raise NotImplementedError

    def _extend_matrix_postprocess(self):
        """Helper method to optionally modify the extended matrix after calling ``self.extend_matrix``."""
        pass


class WeakLabels(WeakLabelsBase):
    """Computes the weak labels of a single-label text classification dataset by applying a given list of rules.

    Args:
        dataset: Name of the dataset to which the rules will be applied.
        rules: A list of rules (labeling functions). They must return a string, or ``None`` in case of abstention.
            If None, we will use the rules of the dataset (Default).
        ids: An optional list of record ids to filter the dataset before applying the rules.
        query: An optional ElasticSearch query with the
            `query string syntax <https://argilla.readthedocs.io/en/stable/guides/queries.html>`_
            to filter the dataset before applying the rules.
        label2int: An optional dict, mapping the labels to integers. Remember that the return type ``None`` means
            abstention (e.g. ``{None: -1}``). By default, we will build a mapping on the fly when applying the rules.

    Raises:
        NoRulesFoundError: When you do not provide rules, and the dataset has no rules either.
        DuplicatedRuleNameError: When you provided multiple rules with the same name.
        NoRecordsFoundError: When the filtered dataset is empty.
        MultiLabelError: When trying to get weak labels for a multi-label text classification task.
        MissingLabelError: When provided with a ``label2int`` dict, and a
            weak label or annotation label is not present in its keys.

    Examples:
        >>> # Get the weak label matrix from a dataset with rules:
        >>> weak_labels = WeakLabels(dataset="my_dataset")
        >>> weak_labels.matrix()
        >>> weak_labels.summary()
        >>>
        >>> # Get the weak label matrix from rules defined in Python:
        >>> def awesome_rule(record: TextClassificationRecord) -> str:
        ...     return "Positive" if "awesome" in record.text else None
        >>> another_rule = Rule(query="good OR best", label="Positive")
        >>> weak_labels = WeakLabels(dataset="my_dataset", rules=[awesome_rule, another_rule])
        >>> weak_labels.matrix()
        >>> weak_labels.summary()
        >>>
        >>> # Use the WeakLabels object with snorkel's LabelModel:
        >>> from snorkel.labeling.model import LabelModel
        >>> label_model = LabelModel()
        >>> label_model.fit(L_train=weak_labels.matrix(has_annotation=False))
        >>> label_model.score(L=weak_labels.matrix(has_annotation=True), Y=weak_labels.annotation())
        >>> label_model.predict(L=weak_labels.matrix(has_annotation=False))
        >>>
        >>> # For a builtin integration with Snorkel, see `argilla.labeling.text_classification.Snorkel`.
    """

    def __init__(
        self,
        dataset: str,
        rules: Optional[List[Callable]] = None,
        ids: Optional[List[Union[int, str]]] = None,
        query: Optional[str] = None,
        label2int: Optional[Dict[Optional[str], int]] = None,
    ):
        super().__init__(dataset=dataset, rules=rules, ids=ids, query=query)

        if self._records[0].multi_label:
            raise MultiLabelError(
                "For multi-label text classification, use the `rg.labeling.text_classification.WeakMultiLabels` class."
            )

        # apply rules -> create the weak label matrix, annotation array, final label2int mapping
        self._matrix, self._annotation, self._label2int = self._apply_rules(label2int)
        self._int2label = {v: k for k, v in self._label2int.items()}

    def _apply_rules(
        self, label2int: Optional[Dict[str, int]]
    ) -> Tuple[np.ndarray, np.ndarray, Dict[Optional[str], int]]:
        """Apply the rules to the dataset.

        Args:
            label2int: An optional custom label2int mapping.

        Returns:
            The weak label matrix, the annotation array and their label2int mapping.

        Raises:
            MissingLabelError: When provided with a ``label2int`` dict, and a
                weak label or annotation label is not present in its keys.
        """
        # call apply on the ElasticSearch rules
        for rule in tqdm(self._rules, desc="Preparing rules"):
            if isinstance(rule, Rule):
                rule.apply(self._dataset)

        # create weak label matrix, annotation array, final label2int
        weak_label_matrix = np.empty((len(self._records), len(self._rules)), dtype=np.short)
        annotation_array = np.empty(len(self._records), dtype=np.short)
        _label2int = {None: -1} if label2int is None else label2int
        if None not in _label2int:
            raise MissingLabelError(
                "Your provided `label2int` mapping does not contain the required abstention label `None`."
            )

        for n, record in tqdm(enumerate(self._records), total=len(self._records), desc="Applying rules"):
            # FIRST: fill annotation array
            try:
                annotation = _label2int[record.annotation]
            except KeyError as error:
                # When a label2int was provided, we want to raise an error if the label is missing!
                if label2int is not None:
                    raise MissingLabelError(
                        f"The annotation label '{record.annotation}' is missing in the `label2int` dict {label2int}"
                    ) from error
                # we already have `None` -> we need to subtract 1
                _label2int[record.annotation] = len(_label2int) - 1
                annotation = _label2int[record.annotation]

            annotation_array[n] = annotation

            # SECOND: fill weak label matrix
            for m, rule in enumerate(self._rules):
                weak_label = rule(record)
                if isinstance(weak_label, list):
                    if len(weak_label) != 1:
                        raise MultiLabelError(
                            "For rules that do not return exactly 1 label, use the `WeakMultiLabels` class."
                        )
                    weak_label = weak_label[0]

                try:
                    weak_label = _label2int[weak_label]
                except KeyError as error:
                    if label2int is not None:
                        raise MissingLabelError(
                            f"A rule returned the weak label '{weak_label}', "
                            f"but it is missing in the `label2int` dict {label2int}"
                        ) from error
                    _label2int[weak_label] = len(_label2int) - 1
                    weak_label = _label2int[weak_label]

                weak_label_matrix[n, m] = weak_label

        return weak_label_matrix, annotation_array, _label2int

    @property
    def cardinality(self) -> int:
        return len(self._label2int) - 1

    @property
    def labels(self) -> List[str]:
        return [key for key in self._label2int.keys() if key is not None]

    @property
    def label2int(self) -> Dict[Optional[str], int]:
        """The dictionary that maps weak/annotation labels to integers."""
        return self._label2int

    @property
    def int2label(self) -> Dict[int, Optional[str]]:
        """The dictionary that maps integers to weak/annotation labels."""
        return self._int2label

    def matrix(self, has_annotation: Optional[bool] = None) -> np.ndarray:
        """Returns the weak label matrix, or optionally just a part of it.

        Args:
            has_annotation: If True, return only the part of the matrix that has a corresponding annotation.
                If False, return only the part of the matrix that has NOT a corresponding annotation.
                By default, we return the whole weak label matrix.

        Returns:
            The weak label matrix, or optionally just a part of it.
        """
        matrix = self._matrix if self._extended_matrix is None else self._extended_matrix

        if has_annotation is True:
            return matrix[self._annotation != self._label2int[None]]
        if has_annotation is False:
            return matrix[self._annotation == self._label2int[None]]

        return matrix

    def annotation(
        self,
        include_missing: bool = False,
        exclude_missing_annotations: Optional[bool] = None,
    ) -> np.ndarray:
        """Returns the annotation labels as an array of integers.

        Args:
            include_missing: If True, returns an array of the length of the record list (``self.records()``).
                For this, we will fill the array with the ``self.label2int[None]`` integer for records without an annotation.
            exclude_missing_annotations: DEPRECATED

        Returns:
            The annotation array of integers.
        """
        if exclude_missing_annotations is not None:
            warnings.warn(
                "'exclude_missing_annotations' is deprecated and will be removed in the next major release. "
                "Please use the 'include_missing' argument.",
                category=FutureWarning,
            )
            include_missing = not exclude_missing_annotations

        if include_missing:
            return self._annotation

        return self._annotation[self._annotation != self._label2int[None]]

    def summary(
        self,
        normalize_by_coverage: bool = False,
        annotation: Optional[np.ndarray] = None,
    ) -> pd.DataFrame:
        """Returns following summary statistics for each rule:

        - **label**: Set of unique labels returned by the rule, excluding "None" (abstain).
        - **coverage**: Fraction of the records labeled by the rule.
        - **annotated_coverage**: Fraction of annotated records labeled by the rule (if annotations are available).
        - **overlaps**: Fraction of the records labeled by the rule together with at least one other rule.
        - **conflicts**: Fraction of the records where the rule disagrees with at least one other rule.
        - **correct**: Number of labels the rule predicted correctly (if annotations are available).
        - **incorrect**: Number of labels the rule predicted incorrectly (if annotations are available).
        - **precision**: Fraction of correct labels given by the rule (if annotations are available). The precision does not penalize the rule for abstains.


        Args:
            normalize_by_coverage: Normalize the overlaps and conflicts by the respective coverage.
            annotation: An optional array with ints holding the annotations.
                By default, we will use ``self.annotation(include_missing=True)``.

        Returns:
            The summary statistics for each rule in a pandas DataFrame.
        """
        annotation = annotation if annotation is not None else self._annotation
        has_weak_label = self.matrix() != self._label2int[None]

        # polarity (label)
        polarity = [
            set(
                self._int2label[integer]
                for integer in np.unique(self.matrix()[:, i][self.matrix()[:, i] != self._label2int[None]])
            )
            for i in range(len(self._rules))
        ]
        polarity.append(set().union(*polarity))

        # coverage
        coverage = has_weak_label.sum(axis=0) / len(self._records)
        coverage = np.append(
            coverage,
            (has_weak_label.sum(axis=1) > 0).sum() / len(self._records),
        )

        # overlaps
        has_overlaps = has_weak_label.sum(axis=1) > 1
        overlaps = self._compute_overlaps_conflicts(has_weak_label, has_overlaps, coverage, normalize_by_coverage)

        # conflicts
        # TODO: For a lot of records (~1e6), this could become slow (~10s) ... a vectorized solution would be better.
        has_conflicts = np.apply_along_axis(
            lambda x: len(np.unique(x[x != self._label2int[None]])) > 1,
            axis=1,
            arr=self.matrix(),
        )
        conflicts = self._compute_overlaps_conflicts(has_weak_label, has_conflicts, coverage, normalize_by_coverage)

        # index for the summary
        index = list(self._rules_name2index.keys()) + ["total"]

        # only add annotated_coverage, correct, incorrect and precision if we have annotations
        has_annotation = annotation != self._label2int[None]
        if any(has_annotation):
            # annotated coverage
            annotated_coverage = has_weak_label[has_annotation].sum(axis=0) / has_annotation.sum()
            annotated_coverage = np.append(
                annotated_coverage,
                (has_weak_label[has_annotation].sum(axis=1) > 0).sum() / has_annotation.sum(),
            )

            # correct/incorrect
            correct, incorrect = self._compute_correct_incorrect(
                has_weak_label,
                annotation if annotation is not None else self._annotation,
            )

            # precision, ignore division by 0 warnings: we allow np.nan and np.inf
            with np.errstate(divide="ignore", invalid="ignore"):
                precision = correct / (correct + incorrect)

            return pd.DataFrame(
                {
                    "label": polarity,
                    "coverage": coverage,
                    "annotated_coverage": annotated_coverage,
                    "overlaps": overlaps,
                    "conflicts": conflicts,
                    "correct": correct,
                    "incorrect": incorrect,
                    "precision": precision,
                },
                index=index,
            )

        return pd.DataFrame(
            {
                "label": polarity,
                "coverage": coverage,
                "overlaps": overlaps,
                "conflicts": conflicts,
            },
            index=index,
        )

    def _compute_correct_incorrect(
        self, has_weak_label: np.ndarray, annotation: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Helper method to compute the correctly and incorrectly predicted annotations by the rules"""
        annotation_matrix = np.repeat(annotation, len(self._rules)).reshape(self.matrix().shape)

        # correct
        correct_with_abstain = annotation_matrix == self.matrix()
        correct = np.where(has_weak_label, correct_with_abstain, False).sum(axis=0)

        # incorrect
        incorrect_with_abstain = annotation_matrix != self.matrix()
        incorrect = np.where(
            has_weak_label & (annotation_matrix != self._label2int[None]),
            incorrect_with_abstain,
            False,
        ).sum(axis=0)

        # add totals at the end
        return np.append(correct, correct.sum()), np.append(incorrect, incorrect.sum())

    def show_records(
        self,
        labels: Optional[List[str]] = None,
        rules: Optional[List[Union[str, int]]] = None,
    ) -> pd.DataFrame:
        """Shows records in a pandas DataFrame, optionally filtered by weak labels and non-abstaining rules.

        If you provide both ``labels`` and ``rules``, we take the intersection of both filters.

        Args:
            labels: All of these labels are in the record's weak labels. If None, do not filter by labels.
            rules: All of these rules did not abstain for the record. If None, do not filter by rules.
                You can refer to the rules by their (function) name or by their index in the ``self.rules`` list.

        Returns:
            The optionally filtered records as a pandas DataFrame.
        """
        # get labels mask
        if labels is not None:
            labels = [self._label2int[label] for label in labels]
            idx_by_labels = np.isin(self.matrix(), labels).sum(axis=1) >= len(labels)
        else:
            idx_by_labels = np.ones_like(self._records).astype(bool)

        # get rule mask
        if rules is not None:
            rules = [self._rules_name2index[rule] if isinstance(rule, str) else rule for rule in rules]
            idx_by_rules = (self.matrix()[:, rules] != self._label2int[None]).sum(axis=1) == len(rules)
        else:
            idx_by_rules = np.ones_like(self._records).astype(bool)

        # apply mask
        filtered_records = np.array(self._records)[idx_by_labels & idx_by_rules]

        return pd.DataFrame(map(lambda x: x.dict(), filtered_records))

    def change_mapping(self, label2int: Dict[str, int]):
        """Allows you to change the mapping between labels and integers.

        This will update the ``self.matrix`` as well as the ``self.annotation``.

        Args:
            label2int: New label to integer mapping. Must cover all previous labels.
        """
        # save masks for swapping
        label_masks = {}
        annotation_masks = {}

        for label in self._label2int:
            # Check new label2int mapping
            if label not in label2int:
                raise MissingLabelError(f"The label '{label}' is missing in the new mapping.")
            # compute masks
            label_masks[label] = self.matrix() == self._label2int[label]
            annotation_masks[label] = self._annotation == self._label2int[label]

        # swap integers
        for label in self._label2int:
            self.matrix()[label_masks[label]] = label2int[label]
            self._annotation[annotation_masks[label]] = label2int[label]

        # update mapping dicts
        self._label2int = label2int.copy()
        self._int2label = {val: key for key, val in self._label2int.items()}

    @_add_docstr(WeakLabelsBase.extend_matrix.__doc__.format(class_name="WeakLabels"))
    def extend_matrix(
        self,
        thresholds: Union[List[float], np.ndarray],
        embeddings: Optional[np.ndarray] = None,
        gpu: bool = False,
    ):
        super().extend_matrix(thresholds=thresholds, embeddings=embeddings, gpu=gpu)

    def _extend_matrix_preprocess(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        abstains = [
            np.argwhere(self._matrix[:, i] == self._label2int[None]).flatten() for i in range(self._matrix.shape[1])
        ]

        supports = [
            np.argwhere(self._matrix[:, i] != self._label2int[None]).flatten() for i in range(self._matrix.shape[1])
        ]

        return abstains, supports

    def _extend_matrix_postprocess(self):
        """Keeps the rows of the original weak label matrix, for which at least on rule did not abstain."""
        recs_with_votes = np.argwhere((self._matrix != self._label2int[None]).sum(-1) > 0).flatten()
        self._extended_matrix[recs_with_votes] = self._matrix[recs_with_votes]


class WeakMultiLabels(WeakLabelsBase):
    """Computes the weak labels of a multi-label text classification dataset by applying a given list of rules.

    Args:
        dataset: Name of the dataset to which the rules will be applied.
        rules: A list of rules (labeling functions). They must return a string, list of strings, or ``None`` in case of
            abstention. If None, we will use the rules of the dataset (Default).
        ids: An optional list of record ids to filter the dataset before applying the rules.
        query: An optional ElasticSearch query with the
            `query string syntax <https://argilla.readthedocs.io/en/stable/guides/queries.html>`_
            to filter the dataset before applying the rules.

    Raises:
        NoRulesFoundError: When you do not provide rules, and the dataset has no rules either.
        DuplicatedRuleNameError: When you provided multiple rules with the same name.
        NoRecordsFoundError: When the filtered dataset is empty.

    Examples:
        >>> # Get the 3 dimensional weak label matrix from a multi-label classification dataset with rules:
        >>> weak_labels = WeakMultiLabels(dataset="my_dataset")
        >>> weak_labels.matrix()
        >>> weak_labels.summary()
        >>>
        >>> # Get the 3 dimensional weak label matrix from rules defined in Python:
        >>> def awesome_rule(record: TextClassificationRecord) -> str:
        ...     return ["Positive", "Slang"] if "next level" in record.text else None
        >>> another_rule = Rule(query="amped OR psyched", label=["Positive", "Slang"])
        >>> weak_labels = WeakMultiLabels(dataset="my_dataset", rules=[awesome_rule, another_rule])
        >>> weak_labels.matrix()
        >>> weak_labels.summary()
    """

    def __init__(
        self,
        dataset: str,
        rules: Optional[List[Callable]] = None,
        ids: Optional[List[Union[int, str]]] = None,
        query: Optional[str] = None,
    ):
        super().__init__(dataset=dataset, rules=rules, ids=ids, query=query)

        # apply rules -> create the weak label matrix (3D), annotation (2D), label list
        self._matrix, self._annotation, self._labels = self._apply_rules()

    def _apply_rules(self) -> Tuple[np.ndarray, np.ndarray, List[str]]:
        # call apply on the ElasticSearch rules
        for rule in tqdm(self._rules, desc="Preparing rules"):
            if isinstance(rule, Rule):
                rule.apply(self._dataset)

        # we make two passes over the records:
        # FIRST: Get labels from rules and annotations
        annotations, weak_labels = [], []
        for record in tqdm(self._records, total=len(self._records), desc="Applying rules"):
            annotations.append(record.annotation if isinstance(record.annotation, list) else [record.annotation])
            weak_labels.append([np.atleast_1d(rule(record)) for rule in self._rules])

        annotation_set = {ann for anns in annotations for ann in anns}
        weak_label_set = {wl for wl_record in weak_labels for wl_rule in wl_record for wl in wl_rule}
        labels = sorted(list(annotation_set.union(weak_label_set) - {None}))

        # create weak label matrix (3D), annotation matrix
        weak_label_matrix = np.empty((len(self._records), len(self._rules), len(labels)), dtype=np.byte)
        annotation_matrix = np.empty((len(self._records), len(labels)), dtype=np.byte)

        # SECOND: Fill arrays with weak labels
        for n, annotation_n, weak_label_n in tqdm(
            zip(range(len(annotations)), annotations, weak_labels),
            desc="Filling weak label matrix",
            total=len(annotations),
        ):
            # first: fill annotation matrix
            if annotation_n == [None]:
                # "abstain" is an array with -1
                annotation_matrix[n] = -1 * np.ones(len(labels), dtype=np.byte)
            else:
                annotation_matrix[n] = np.array(
                    [1 if label in annotation_n else 0 for label in labels],
                    dtype=np.byte,
                )

            # second: fill weak label matrix (3D)
            for m, weak_labels_m in enumerate(weak_label_n):
                if weak_labels_m.tolist() == [None]:
                    weak_label_matrix[n, m] = -1 * np.ones(len(labels))
                else:
                    weak_label_matrix[n, m] = np.array(
                        [1 if label in weak_labels_m else 0 for label in labels],
                        dtype=np.byte,
                    )

        return weak_label_matrix, annotation_matrix, labels

    @property
    def labels(self) -> List[str]:
        """The labels of the multi-label text classification dataset."""
        return self._labels

    @property
    def cardinality(self) -> int:
        return len(self._labels)

    def matrix(self, has_annotation: Optional[bool] = None) -> np.ndarray:
        """Returns the 3 dimensional weak label matrix, or optionally just a part of it.

        It has the dimensions ("nr of record" x "nr of rules" x "nr of labels").
        It holds a 1 or 0 in case a rule votes for a label or not. If the rule abstains, it holds a -1 for each label.

        Args:
            has_annotation: If True, return only the part of the matrix that has a corresponding annotation.
                If False, return only the part of the matrix that has NOT a corresponding annotation.
                By default, we return the whole weak label matrix.

        Returns:
            The 3 dimensional weak label matrix, or optionally just a part of it.
        """
        matrix = self._matrix if self._extended_matrix is None else self._extended_matrix

        if has_annotation is True:
            return matrix[self._annotation.sum(1) >= 0]
        if has_annotation is False:
            return matrix[self._annotation.sum(1) < 0]

        return matrix

    def annotation(
        self,
        include_missing: bool = False,
    ) -> np.ndarray:
        """Returns the annotation labels as a matrix of integers.

        It has the dimensions ("nr of record" x "nr of labels").
        It holds a 1 or 0 to indicate if the record is annotated with the corresponding label.
        In case there is no annotation for the record, it holds a -1 for each label.

        Args:
            include_missing: If True, returns a matrix of the length of the record list (``self.records()``).
                For this, we will fill the matrix with -1 for records without an annotation.

        Returns:
            The annotation labels as a matrix of integers.
        """
        if include_missing:
            return self._annotation

        return self._annotation[self._annotation.sum(1) >= 0]

    def summary(
        self,
        normalize_by_coverage: bool = False,
        annotation: Optional[np.ndarray] = None,
    ) -> pd.DataFrame:
        """Returns following summary statistics for each rule:

        - **label**: Set of unique labels returned by the rule, excluding "None" (abstain).
        - **coverage**: Fraction of the records labeled by the rule.
        - **annotated_coverage**: Fraction of annotated records labeled by the rule (if annotations are available).
        - **overlaps**: Fraction of the records labeled by the rule together with at least one other rule.
        - **correct**: Number of labels the rule predicted correctly (if annotations are available).
        - **incorrect**: Number of labels the rule predicted incorrectly or missed (if annotations are available).
        - **precision**: Fraction of correct labels given by the rule (if annotations are available). The precision does not penalize the rule for abstains.

        Args:
            normalize_by_coverage: Normalize the overlaps by the respective coverage.
            annotation: An optional matrix with ints holding the annotations (see ``self.annotation``).
                By default, we will use ``self.annotation(include_missing=True)``.

        Returns:
            The summary statistics for each rule in a pandas DataFrame.
        """
        annotation = annotation if annotation is not None else self._annotation
        has_weak_label = self.matrix().sum(2) >= 0

        # polarity (label)
        polarity = [
            set(
                [
                    self._labels[i]
                    # get indices of votes
                    for i in np.nonzero(
                        # remove abstentions
                        self.matrix()[:, m, :][self.matrix()[:, m, :].sum(1) >= 0]
                    )[1]
                ]
            )
            for m in range(len(self._rules))
        ]
        polarity.append(set().union(*polarity))

        # coverage
        coverage = has_weak_label.sum(axis=0) / len(self._records)
        coverage = np.append(
            coverage,
            (has_weak_label.sum(axis=1) > 0).sum() / len(self._records),
        )

        # overlaps
        has_overlaps = has_weak_label.sum(axis=1) > 1
        overlaps = self._compute_overlaps_conflicts(has_weak_label, has_overlaps, coverage, normalize_by_coverage)

        # index for the summary
        index = list(self._rules_name2index.keys()) + ["total"]

        # only add annotated_coverage, correct, incorrect and precision if we have annotations
        has_annotation = annotation.sum(1) >= 0
        if any(has_annotation):
            # annotated coverage
            annotated_coverage = has_weak_label[has_annotation].sum(axis=0) / has_annotation.sum()
            annotated_coverage = np.append(
                annotated_coverage,
                (has_weak_label[has_annotation].sum(axis=1) > 0).sum() / has_annotation.sum(),
            )

            # correct/incorrect
            correct, incorrect = self._compute_correct_incorrect(annotation)

            # precision, ignore division by 0 warnings: we allow np.nan and np.inf
            with np.errstate(divide="ignore", invalid="ignore"):
                precision = correct / (correct + incorrect)

            return pd.DataFrame(
                {
                    "label": polarity,
                    "coverage": coverage,
                    "annotated_coverage": annotated_coverage,
                    "overlaps": overlaps,
                    "correct": correct,
                    "incorrect": incorrect,
                    "precision": precision,
                },
                index=index,
            )

        return pd.DataFrame(
            {
                "label": polarity,
                "coverage": coverage,
                "overlaps": overlaps,
            },
            index=index,
        )

    def _compute_correct_incorrect(self, annotation: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Helper method to compute the correctly and incorrectly predicted annotations by the rules"""
        # transform annotation to tensor
        annotation = np.repeat(annotation, len(self._rules), axis=0).reshape(self.matrix().shape)

        # correct, we don't want to count the "correct non predictions"
        correct = ((annotation == self.matrix()) & (self.matrix() == 1)).sum(2).sum(0)

        # incorrect, we don't want to count the "misses", since we focus on precision, not recall
        incorrect = ((annotation != self.matrix()) & (self.matrix() == 1) & (annotation != -1)).sum(2).sum(0)

        # add totals at the end
        return np.append(correct, correct.sum()), np.append(incorrect, incorrect.sum())

    def show_records(
        self,
        labels: Optional[List[str]] = None,
        rules: Optional[List[Union[str, int]]] = None,
    ) -> pd.DataFrame:
        """Shows records in a pandas DataFrame, optionally filtered by weak labels and non-abstaining rules.

        If you provide both ``labels`` and ``rules``, we take the intersection of both filters.

        Args:
            labels: All of these labels are in the record's weak labels. If None, do not filter by labels.
            rules: All of these rules did not abstain for the record. If None, do not filter by rules.
                You can refer to the rules by their (function) name or by their index in the ``self.rules`` list.

        Returns:
            The optionally filtered records as a pandas DataFrame.
        """
        # get labels mask
        if labels is not None:
            labels = [self._labels.index(label) for label in labels]
            idx_by_labels = np.all(((self.matrix() == 1).sum(1) > 0)[:, labels], axis=1)
        else:
            idx_by_labels = np.ones_like(self._records).astype(bool)

        # get rule mask
        if rules is not None:
            rules = [self._rules_name2index[rule] if isinstance(rule, str) else rule for rule in rules]
            idx_by_rules = (self.matrix()[:, rules, :].sum(axis=2) >= 0).sum(axis=1) == len(rules)
        else:
            idx_by_rules = np.ones_like(self._records).astype(bool)

        # apply mask
        filtered_records = np.array(self._records)[idx_by_labels & idx_by_rules]

        return pd.DataFrame(map(lambda x: x.dict(), filtered_records))

    @_add_docstr(WeakLabelsBase.extend_matrix.__doc__.format(class_name="WeakMultiLabels"))
    def extend_matrix(
        self,
        thresholds: Union[List[float], np.ndarray],
        embeddings: Optional[np.ndarray] = None,
        gpu: bool = False,
    ):
        super().extend_matrix(thresholds=thresholds, embeddings=embeddings, gpu=gpu)

    def _extend_matrix_preprocess(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        abstains = [np.argwhere(self._matrix[:, i].sum(-1) < 0).flatten() for i in range(self._matrix.shape[1])]

        supports = [np.argwhere(self._matrix[:, i].sum(-1) >= 0).flatten() for i in range(self._matrix.shape[1])]

        return abstains, supports


class WeakLabelsError(Exception):
    pass


class NoRulesFoundError(WeakLabelsError):
    pass


class DuplicatedRuleNameError(WeakLabelsError):
    pass


class NoRecordsFoundError(WeakLabelsError):
    pass


class MultiLabelError(WeakLabelsError):
    pass


class MissingLabelError(WeakLabelsError):
    pass



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel

from argilla.client.sdk.users.models import UserRole


class WorkspaceModel(BaseModel):
    id: UUID
    name: str
    inserted_at: datetime
    updated_at: datetime


# TODO(alvarobartt): replace with the `pydantic.BaseModel` created at https://github.com/argilla-io/argilla/pull/3169 once is merged
class WorkspaceUserModel(BaseModel):
    id: UUID
    first_name: str
    last_name: Optional[str]
    full_name: Optional[str]
    username: str
    role: UserRole
    workspaces: Optional[List[str]]
    api_key: str
    inserted_at: datetime
    updated_at: datetime



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

stopwords_data = """
 | From https://snowballstem.org/algorithms/english/stop.txt
 | This file is distributed under the BSD License.
 | See https://snowballstem.org/license.html
 | Also see https://opensource.org/licenses/bsd-license.html
 |  - Encoding was converted to UTF-8.
 |  - This notice was added.
 |
 | NOTE: To use this file with StopFilterFactory, you must specify format="snowball"

 | An English stop word list. Comments begin with vertical bar. Each stop
 | word is at the start of a line.

 | Many of the forms below are quite rare (e.g. "yourselves") but included for
 |  completeness.

           | PRONOUNS FORMS
             | 1st person sing

i              | subject, always in upper case of course

me             | object
my             | possessive adjective
               | the possessive pronoun `mine' is best suppressed, because of the
               | sense of coal-mine etc.
myself         | reflexive
             | 1st person plural
we             | subject

| us           | object
               | care is required here because US = United States. It is usually
               | safe to remove it if it is in lower case.
our            | possessive adjective
ours           | possessive pronoun
ourselves      | reflexive
             | second person (archaic `thou' forms not included)
you            | subject and object
your           | possessive adjective
yours          | possessive pronoun
yourself       | reflexive (singular)
yourselves     | reflexive (plural)
             | third person singular
he             | subject
him            | object
his            | possessive adjective and pronoun
himself        | reflexive

she            | subject
her            | object and possessive adjective
hers           | possessive pronoun
herself        | reflexive

it             | subject and object
its            | possessive adjective
itself         | reflexive
             | third person plural
they           | subject
them           | object
their          | possessive adjective
theirs         | possessive pronoun
themselves     | reflexive
             | other forms (demonstratives, interrogatives)
what
which
who
whom
this
that
these
those

            | VERB FORMS (using F.R. Palmer's nomenclature)
             | BE
am             | 1st person, present
is             | -s form (3rd person, present)
are            | present
was            | 1st person, past
were           | past
be             | infinitive
been           | past participle
being          | -ing form
             | HAVE
have           | simple
has            | -s form
had            | past
having         | -ing form
             | DO
do             | simple
does           | -s form
did            | past
doing          | -ing form

 | The forms below are, I believe, best omitted, because of the significant
 | homonym forms:

 |  He made a WILL
 |  old tin CAN
 |  merry month of MAY
 |  a smell of MUST
 |  fight the good fight with all thy MIGHT

 | would, could, should, ought might however be included

 |          | AUXILIARIES
 |            | WILL
 |will

would

 |            | SHALL
 |shall

should

 |            | CAN
 |can

could

 |            | MAY
 |may
 |might
 |            | MUST
 |must
 |            | OUGHT

ought

           | COMPOUND FORMS, increasingly encountered nowadays in 'formal' writing
              | pronoun + verb

i'm
you're
he's
she's
it's
we're
they're
i've
you've
we've
they've
i'd
you'd
he'd
she'd
we'd
they'd
i'll
you'll
he'll
she'll
we'll
they'll

              | verb + negation

isn't
aren't
wasn't
weren't
hasn't
haven't
hadn't
doesn't
don't
didn't

              | auxiliary + negation

won't
wouldn't
shan't
shouldn't
can't
cannot
couldn't
mustn't

             | miscellaneous forms

let's
that's
who's
what's
here's
there's
when's
where's
why's
how's

              | rarer forms

 | daren't needn't

              | doubtful forms

 | oughtn't mightn't

           | ARTICLES
a
an
the

           | THE REST (Overlap among prepositions, conjunctions, adverbs etc is so
           | high, that classification is pointless.)
and
but
if
or
because
as
until
while

of
at
by
for
with
about
against
between
into
through
during
before
after
above
below
to
from
up
down
in
out
on
off
over
under

again
further
then
once

here
there
when
where
why
how

all
any
both
each
few
more
most
other
some
such

no
nor
not
only
own
same
so
than
too
very

| Just for the record, the following words are among the commonest in English

    | one
    | every
    | least
    | less
    | many
    | now
    | ever
    | never
    | say
    | says
    | said
    | also
    | get
    | go
    | goes
    | just
    | made
    | make
    | put
    | see
    | seen
    | whether
    | like
    | well
    | back
    | even
    | still
    | way
    | take
    | since
    | another
    | however
    | two
    | three
    | four
    | five
    | first
    | second
    | new
    | old
    | high
    | long

"""


def read_stopwords(data: str):
    stopwords = []
    for line in data.split("\n"):
        word = line.split("|")[0].strip()
        if word:
            stopwords.append(word)
    return list(set(stopwords))


STOPWORDS = read_stopwords(stopwords_data)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pydantic import BaseSettings

from argilla.server import helpers
from argilla.server.settings import settings as server_settings


class Settings(BaseSettings):

    """
    Attributes
    ----------

    secret_key:
        The secret key used for signed the token data

    algorithm:
        Encryption algorithm for token data

    token_expiration_in_minutes:
        The session token expiration in minutes. Default=30000

    """

    secret_key: str = "secret"
    algorithm: str = "HS256"
    token_expiration_in_minutes: int = 30000
    token_api_url: str = "/api/security/token"
    users_db_file: str = ".users.yml"

    @property
    def public_oauth_token_url(self):
        """The final public token url used for openapi doc setup"""
        return f"{server_settings.base_url}{helpers.remove_prefix(self.token_api_url,'/')}"

    class Config:
        env_prefix = "ARGILLA_LOCAL_AUTH_"

        fields = {
            "secret_key": {"env": ["SECRET_KEY", f"{env_prefix}SECRET_KEY"]},
            "token_expiration_in_minutes": {
                "env": [
                    "TOKEN_EXPIRATION_IN_MINUTES",
                    f"{env_prefix}TOKEN_EXPIRATION_IN_MINUTES",
                ]
            },
        }


settings = Settings()



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Union

from pydantic import BaseModel, Field, root_validator, validator

from argilla._constants import DEFAULT_MAX_KEYWORD_LENGTH
from argilla.server.commons.models import PredictionStatus, TaskStatus, TaskType
from argilla.server.helpers import flatten_dict
from argilla.server.services.datasets import ServiceBaseDataset
from argilla.server.services.search.model import ServiceBaseRecordsQuery, ServiceScoreRange
from argilla.server.services.tasks.commons import ServiceBaseAnnotation, ServiceBaseRecord


class ServiceLabelingRule(BaseModel):
    query: str = Field(description="The es rule query")

    author: str = Field(description="User who created the rule")
    created_at: Optional[datetime] = Field(default_factory=datetime.utcnow, description="Rule creation timestamp")

    label: Optional[str] = Field(default=None, description="@Deprecated::The label associated with the rule.")
    labels: List[str] = Field(
        default_factory=list,
        description="For multi label problems, a list of labels. " "It will replace the `label` field",
    )
    description: Optional[str] = Field(None, description="A brief description of the rule")

    @root_validator
    def initialize_labels(cls, values):
        label = values.get("label", None)
        labels = values.get("labels", [])

        if label:
            labels.append(label)
            values["labels"] = list(set(labels))

        assert len(labels) >= 1, f"No labels was provided in rule {values}"
        return values

    @validator("query")
    def strip_query(cls, query: str) -> str:
        """Remove blank spaces for query"""
        return query.strip()


class ServiceTextClassificationDataset(ServiceBaseDataset):
    task: TaskType = Field(default=TaskType.text_classification)
    rules: List[ServiceLabelingRule] = Field(default_factory=list)


class ClassPrediction(BaseModel):
    """
    Single class prediction

    Attributes:
    -----------

    class_label: Union[str, int]
        the predicted class

    score: float
        the predicted class score. For human-supervised annotations,
        this probability should be 1.0
    """

    class_label: Union[str, int] = Field(alias="class")
    score: float = Field(default=1.0, ge=0.0, le=1.0)

    @validator("class_label")
    def check_label_length(cls, class_label):
        if isinstance(class_label, str):
            assert 1 <= len(class_label) <= DEFAULT_MAX_KEYWORD_LENGTH, (
                f"Class name '{class_label}' exceeds max length of {DEFAULT_MAX_KEYWORD_LENGTH}"
                if len(class_label) > DEFAULT_MAX_KEYWORD_LENGTH
                else "Class name must not be empty"
            )
        return class_label

    # See <https://pydantic-docs.helpmanual.io/usage/model_config>
    class Config:
        allow_population_by_field_name = True


class LabelingRuleMetricsSummary(BaseModel):
    """Metrics generated for a labeling rule"""

    coverage: Optional[float] = None
    coverage_annotated: Optional[float] = None
    correct: Optional[float] = None
    incorrect: Optional[float] = None
    precision: Optional[float] = None

    total_records: int
    annotated_records: int


class DatasetLabelingRulesMetricsSummary(BaseModel):
    coverage: Optional[float] = None
    coverage_annotated: Optional[float] = None

    total_records: int
    annotated_records: int


class TextClassificationAnnotation(ServiceBaseAnnotation):
    """
    Annotation class for text classification tasks

    Attributes:
    -----------

    labels: List[LabelPrediction]
        list of annotated labels with score
    """

    # TODO(@frascuchon): labels must be a dict (to avoid repeat labels)
    labels: List[ClassPrediction]

    @validator("labels")
    def sort_labels(cls, labels: List[ClassPrediction]):
        """Sort provided labels by score"""
        return sorted(labels, key=lambda x: x.score, reverse=True)


class TokenAttributions(BaseModel):
    """
    The token attributions explaining predicted labels

    Attributes:
    -----------

    token: str
        The input token
    attributions: Dict[str, float]
        A dictionary containing label class-attribution pairs

    """

    token: str
    attributions: Dict[str, float] = Field(default_factory=dict)


class ServiceTextClassificationRecord(ServiceBaseRecord[TextClassificationAnnotation]):
    inputs: Dict[str, Union[str, List[str]]]
    multi_label: bool = False
    explanation: Optional[Dict[str, List[TokenAttributions]]] = None

    class Config:
        allow_population_by_field_name = True

    _SCORE_DEVIATION_ERROR: ClassVar[float] = 0.001

    @root_validator
    def validate_record(cls, values):
        """fastapi validator method"""
        prediction = values.get("prediction", None)
        annotation = values.get("annotation", None)
        status = values.get("status")
        multi_label = values.get("multi_label", False)

        cls._check_score_integrity(prediction, multi_label)
        cls._check_annotation_integrity(annotation, multi_label, status)

        return values

    @classmethod
    def _check_annotation_integrity(
        cls,
        annotation: TextClassificationAnnotation,
        multi_label: bool,
        status: TaskStatus,
    ):
        if status == TaskStatus.validated and not multi_label:
            assert annotation and len(annotation.labels) > 0, "Annotation must include some label for validated records"

        if not multi_label and annotation:
            assert len(annotation.labels) == 1, "Single label record must include only one annotation label"

    @classmethod
    def _check_score_integrity(cls, prediction: TextClassificationAnnotation, multi_label: bool):
        """
        Checks the score value integrity

        Parameters
        ----------
        prediction:
            The prediction annotation
        multi_label:
            If multi label

        """
        if prediction and not multi_label:
            assert sum([label.score for label in prediction.labels]) <= (
                1.0 + cls._SCORE_DEVIATION_ERROR
            ), f"Wrong score distributions: {prediction.labels}"

    @classmethod
    def task(cls) -> TaskType:
        """The task type"""
        return TaskType.text_classification

    @property
    def predicted(self) -> Optional[PredictionStatus]:
        if self.predicted_by and self.annotated_by:
            return PredictionStatus.OK if set(self.predicted_as) == set(self.annotated_as) else PredictionStatus.KO
        return None

    @property
    def predicted_as(self) -> List[str]:
        return self._labels_from_annotation(self.prediction, multi_label=self.multi_label)

    @property
    def annotated_as(self) -> List[str]:
        return self._labels_from_annotation(self.annotation, multi_label=self.multi_label)

    @property
    def scores(self) -> List[float]:
        if not self.prediction:
            return []
        return (
            [label.score for label in self.prediction.labels]
            if self.multi_label
            else [
                prediction_class.score
                for prediction_class in [self._max_class_prediction(self.prediction, multi_label=self.multi_label)]
                if prediction_class
            ]
        )

    def all_text(self) -> str:
        sentences = []
        for v in self.inputs.values():
            if isinstance(v, list):
                sentences.extend(v)
            else:
                sentences.append(v)
        return "\n".join(sentences)

    @validator("inputs")
    def validate_inputs(cls, text: Dict[str, Any]):
        assert len(text) > 0, "No inputs provided"

        for t in text.values():
            assert t is not None, "Cannot include None fields"

        return text

    @validator("inputs")
    def flatten_text(cls, text: Dict[str, Any]):
        flat_dict = flatten_dict(text)
        return flat_dict

    @classmethod
    def _labels_from_annotation(
        cls, annotation: TextClassificationAnnotation, multi_label: bool
    ) -> Union[List[str], List[int]]:
        if not annotation:
            return []

        if multi_label:
            return [label.class_label for label in annotation.labels if label.score > 0.5]

        class_prediction = cls._max_class_prediction(annotation, multi_label=multi_label)
        if class_prediction is None:
            return []

        return [class_prediction.class_label]

    @staticmethod
    def _max_class_prediction(p: TextClassificationAnnotation, multi_label: bool) -> Optional[ClassPrediction]:
        if multi_label or p is None or not p.labels:
            return None
        return p.labels[0]

    def extended_fields(self) -> Dict[str, Any]:
        words = self.all_text()
        return {
            **super().extended_fields(),
            "text": words,
        }


class ServiceTextClassificationQuery(ServiceBaseRecordsQuery):
    predicted_as: List[str] = Field(default_factory=list)
    annotated_as: List[str] = Field(default_factory=list)
    score: Optional[ServiceScoreRange] = Field(default=None)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)

    uncovered_by_rules: List[str] = Field(default_factory=list)


class DatasetLabelingRulesSummary(BaseModel):
    covered_records: int
    annotated_covered_records: int


class LabelingRuleSummary(BaseModel):
    covered_records: int
    annotated_covered_records: int
    correct_records: int = Field(default=0)
    incorrect_records: int = Field(default=0)
    precision: Optional[float] = None



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This file reflects the user facing API.
If you want to add something here, remember to add it as normal import in the _TYPE_CHECKING section (for IDEs),
as well as in the `_import_structure` dictionary.
"""

import sys as _sys
from typing import TYPE_CHECKING as _TYPE_CHECKING

from argilla.logging import configure_logging as _configure_logging

from . import _version
from .utils import LazyargillaModule as _LazyargillaModule

try:
    from rich.traceback import install as _install_rich

    # Rely on `rich` for tracebacks
    _install_rich()
except ModuleNotFoundError:
    pass

__version__ = _version.version

if _TYPE_CHECKING:
    from argilla.client.api import (
        active_client,
        copy,
        delete,
        delete_records,
        get_workspace,
        init,
        list_datasets,
        list_workspaces,
        load,
        log,
        log_async,
        set_workspace,
    )
    from argilla.client.datasets import (
        DatasetForText2Text,
        DatasetForTextClassification,
        DatasetForTokenClassification,
        read_datasets,
        read_pandas,
    )
    from argilla.client.models import (
        Text2TextRecord,
        TextClassificationRecord,
        TextGenerationRecord,  # TODO Remove TextGenerationRecord
        TokenAttributions,
        TokenClassificationRecord,
    )
    from argilla.client.users import User
    from argilla.client.workspaces import Workspace
    from argilla.datasets import (
        TextClassificationSettings,
        TokenClassificationSettings,
        configure_dataset,
        configure_dataset_settings,
        load_dataset_settings,
    )
    from argilla.feedback import (
        FeedbackDataset,
        FeedbackRecord,
        LabelQuestion,
        MultiLabelQuestion,
        RankingQuestion,
        RatingQuestion,
        ResponseSchema,
        TextField,
        TextQuestion,
        ValueSchema,
    )
    from argilla.listeners import Metrics, RGListenerContext, Search, listener
    from argilla.monitoring.model_monitor import monitor
    from argilla.server.server import app


# TODO: remove me
_import_structure = {
    "feedback": [
        "ArgillaTrainer",
        "LabelQuestionStrategy",
        "MultiLabelQuestionStrategy",
        "RatingQuestionStrategy",
        "FeedbackDataset",
        "FeedbackRecord",
        "LabelQuestion",
        "MultiLabelQuestion",
        "RatingQuestion",
        "RankingQuestion",
        "ResponseSchema",
        "TextField",
        "TextQuestion",
        "ValueSchema",
    ],
    "client.api": [
        "copy",
        "delete",
        "delete_records",
        "get_workspace",
        "init",
        "load",
        "log",
        "log_async",
        "set_workspace",
        "active_client",
        "list_datasets",
        "list_workspaces",
    ],
    "client.models": [
        "Text2TextRecord",
        "TextGenerationRecord",  # TODO Remove TextGenerationRecord
        "TextClassificationRecord",
        "TokenClassificationRecord",
        "TokenAttributions",
    ],
    "client.datasets": [
        "DatasetForText2Text",
        "DatasetForTextClassification",
        "DatasetForTokenClassification",
        "read_datasets",
        "read_pandas",
    ],
    "client.users": ["User"],
    "client.workspaces": ["Workspace"],
    "monitoring.model_monitor": ["monitor"],
    "listeners.listener": [
        "listener",
        "RGListenerContext",
        "Search",
        "Metrics",
    ],
    "datasets": [
        "configure_dataset",
        "load_dataset_settings",
        "configure_dataset_settings",
        "TextClassificationSettings",
        "TokenClassificationSettings",
    ],
    "server.app": ["app"],
}

_sys.modules[__name__] = _LazyargillaModule(
    __name__,
    globals()["__file__"],
    _import_structure,
    module_spec=__spec__,
    extra_objects={"__version__": __version__},
)

_configure_logging()



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import datetime
import json
import logging
import re
from typing import Any, Callable, Dict, List, Optional, Tuple

from argilla.client.models import Record, TextClassificationRecord, TokenClassificationRecord
from argilla.monitoring.base import BaseMonitor
from argilla.utils.dependency import require_version

require_version("starlette>=0.13.0")
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse, Response, StreamingResponse
from starlette.types import Message, Receive

_logger = logging.getLogger(__name__)
_default_tokenization_pattern = re.compile(r"\W+")


def token_classification_mapper(inputs, outputs):
    if isinstance(inputs, str):
        text = inputs
    elif isinstance(inputs, dict):
        text = inputs.get("text", "")
    else:
        text = ""
    tokens = outputs.get("tokens") if isinstance(outputs, dict) else None
    return TokenClassificationRecord(
        text=text,
        tokens=tokens or _default_tokenization_pattern.split(text),
        prediction=[
            (entity["label"], entity["start"], entity["end"])
            for entity in (outputs.get("entities") if isinstance(outputs, dict) else outputs)
        ],
        event_timestamp=datetime.datetime.now(),
    )


def text_classification_mapper(inputs, outputs):
    return TextClassificationRecord(
        inputs=inputs,
        prediction=[(label, score) for label, score in zip(outputs.get("labels", []), outputs.get("scores", []))],
        event_timestamp=datetime.datetime.now(),
    )


class CachedJsonRequest(Request):
    """
    We must a cached version of incoming requests since request body cannot be read from middleware directly.
    See <https://github.com/encode/starlette/issues/847> for more information

    TODO Remove usage of CachedRequest when https://github.com/encode/starlette/pull/848 is released
    """

    @property
    def receive(self) -> Receive:
        body = None
        if hasattr(self, "_body"):
            body = self._body
        if body is not None:

            async def cached_receive() -> Message:
                return dict(type="http.request", body=body)

            return cached_receive
        return self._receive


class ArgillaLogHTTPMiddleware(BaseHTTPMiddleware):
    """An standard starlette middleware that enables argilla logs for http prediction requests"""

    def __init__(
        self,
        api_endpoint: str,
        dataset: str,
        records_mapper: Optional[Callable[[dict, dict], Record]],
        sample_rate: float = 1.0,
        log_interval: float = 1.0,
        agent: Optional[str] = None,
        tags: Dict[str, str] = None,
        *args,
        **kwargs,
    ):
        BaseHTTPMiddleware.__init__(self, *args, **kwargs)

        self._endpoint = api_endpoint
        self._dataset = dataset
        self._records_mapper = records_mapper
        self._monitor_cfg = dict(
            dataset=dataset,
            sample_rate=sample_rate,
            log_interval=log_interval,
            agent=agent,
            tags=tags,
        )
        self._monitor: Optional[BaseMonitor] = None

    def init(self):
        if self._monitor:
            return
        from argilla.client.api import active_api

        self._monitor = BaseMonitor(
            self,
            api=active_api(),
            **self._monitor_cfg,
        )
        self._monitor._prepare_log_data = self._prepare_argilla_data

    async def dispatch(
        self,
        request: Request,
        call_next: RequestResponseEndpoint,
    ) -> Response:
        self.init()

        if self._endpoint != request.url.path:  # Filtering endpoint path
            return await call_next(request)

        cached_request = CachedJsonRequest(
            scope=request.scope,
            receive=request.receive,
            send=request._send,
        )

        try:
            # Must obtain input parameters from request
            if cached_request.method in ["POST", "PUT"]:
                content_type = request.headers.get("Content-type", None)
                if content_type is None:
                    if "application/json" not in content_type:
                        return await call_next(request)
                inputs = await cached_request.json()
            elif cached_request.method == "GET":
                inputs = cached_request.query_params._dict
            else:
                raise NotImplementedError("Only request methods POST, PUT and GET are implemented.")

            # Must obtain response from request
            response: Response = await call_next(cached_request)
            if not isinstance(response, (JSONResponse, StreamingResponse)) or response.status_code >= 400:
                return response

            new_response, outputs = await self._extract_response_content(response)
            self._monitor.send_records(inputs=inputs, outputs=outputs)
            return new_response
        except Exception as ex:
            _logger.error("Cannot log to argilla", exc_info=ex)
            return await call_next(request)

    async def _extract_response_content(self, response: Response) -> Tuple[Response, List[Dict[str, Any]]]:
        """Extracts response body content from response and returns a new processable response"""
        body = b""
        new_response = response
        if isinstance(response, StreamingResponse):
            async for chunk in response.body_iterator:
                body += chunk
            new_response = StreamingResponse(
                content=(chunk for chunk in [body]),
                status_code=response.status_code,
                headers={k: v for k, v in response.headers.items()},
                media_type=response.media_type,
                background=response.background,
            )
        else:
            body = response.body
        return new_response, json.loads(body)

    def _prepare_argilla_data(self, inputs: List[Dict[str, Any]], outputs: List[Dict[str, Any]], **tags):
        # using the base monitor, we only need to provide the input data to the rg.log function
        # and the monitor will handle the sample rate, queue and argilla interaction
        try:
            records = self._records_mapper(inputs, outputs)
            assert records, ValueError("The records_mapper returns and empty record list.")
            if not isinstance(records, list):
                records = [records]
        except Exception as ex:
            records = []
            _logger.error("Cannot log to argilla. Error in records mapper.", exc_info=ex)

        for record in records:
            if self._monitor.agent is not None and not record.prediction_agent:
                record.prediction_agent = self._monitor.agent

        return dict(
            records=records or [],
            name=self._dataset,
            tags=tags,
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, Union

from argilla.client.api import get_workspace, load
from argilla.client.datasets import DatasetForText2Text, DatasetForTextClassification, DatasetForTokenClassification
from argilla.client.models import Framework, Text2TextRecord, TextClassificationRecord, TokenClassificationRecord
from argilla.datasets import TextClassificationSettings, TokenClassificationSettings, load_dataset_settings
from argilla.utils.telemetry import get_telemetry_client

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

if TYPE_CHECKING:
    import spacy


class ArgillaTrainer(object):
    _logger = logging.getLogger("ArgillaTrainer")
    _logger.setLevel(logging.INFO)
    _CLIENT = get_telemetry_client()

    def __init__(
        self,
        name: str,
        framework: str,
        workspace: str = None,
        lang: Optional["spacy.Language"] = None,
        model: Optional[str] = None,
        train_size: Optional[float] = None,
        seed: Optional[int] = None,
        gpu_id: Optional[int] = -1,
        framework_kwargs: Optional[dict] = {},
        **load_kwargs: Optional[dict],
    ) -> None:
        """
        Initialize an Argilla Trainer.

        Args:
            name (str): the name of the dataset you want to load.
            framework (str):
                the framework to use for training. Currently, only "transformers", "setfit", and "spacy"
                are supported.
            lang (spacy.Language):
                the spaCy language model to use for training, just required when `framework="spacy"`.
                Defaults to None, but it will be set to `spacy.blank("en")` if not specified.
            model (str):
                name or path to the baseline model to be used. If not specified will set to a good default
                per framework, if applicable. Defaults to None.
            train_size (float):
                the size of the training set. If not specified, the entire dataset will be used for training,
                which may be an issue if `framework="spacy"` as it requires a validation set. Defaults to None.
            seed (int): the random seed to ensure reproducibility. Defaults to None.
            gpu_id (int):
                the GPU ID to use when training a SpaCy model. Defaults to -1, which means that the CPU
                will be used by default. GPU IDs start in 0, which stands for the default GPU in the system,
                if available.
            framework_kwargs (dict): additional arguments for the framework.
            **load_kwargs: arguments for the rg.load() function.
        """
        self._name = name
        self._workspace = workspace or get_workspace()
        self._multi_label = False
        self._split_applied = False
        self._train_size = train_size
        self._seed = seed  # split is used for train-test-split and should therefore be fixed
        self.model = model

        if train_size:
            self._split_applied = True

        self.rg_dataset_snapshot = load(name=self._name, limit=1, workspace=workspace)
        if not len(self.rg_dataset_snapshot) > 0:
            raise ValueError(f"Dataset {self._name} is empty")

        if isinstance(self.rg_dataset_snapshot, DatasetForTextClassification):
            self._rg_dataset_type = DatasetForTextClassification
            self._multi_label = self.rg_dataset_snapshot[0].multi_label
        elif isinstance(self.rg_dataset_snapshot, DatasetForTokenClassification):
            self._rg_dataset_type = DatasetForTokenClassification
        elif isinstance(self.rg_dataset_snapshot, DatasetForText2Text):
            self._rg_dataset_type = DatasetForText2Text
        else:
            raise NotImplementedError(f"Dataset type {type(self.rg_dataset_snapshot)} is not supported.")

        self.dataset_full = load(name=self._name, **load_kwargs)

        # settings for the dataset
        self._settings = load_dataset_settings(name=self._name, workspace=workspace)
        if self._rg_dataset_type in [DatasetForTextClassification, DatasetForTokenClassification]:
            if self._settings is None:
                self._settings = self.dataset_full._infer_settings_from_records()

        framework = Framework(framework)
        if framework in [Framework.SPACY, Framework.SPACY_TRANSFORMERS]:
            import spacy

            self.dataset_full_prepared = self.dataset_full.prepare_for_training(
                framework=framework,
                settings=self._settings,
                train_size=self._train_size,
                seed=self._seed,
                lang=spacy.blank("en") if lang is None else lang,
            )
        else:
            self.dataset_full_prepared = self.dataset_full.prepare_for_training(
                framework=framework,
                settings=self._settings,
                train_size=self._train_size,
                seed=self._seed,
            )

        if framework is Framework.SETFIT:
            from argilla.training.setfit import ArgillaSetFitTrainer

            self._trainer = ArgillaSetFitTrainer(
                name=self._name,
                workspace=self._workspace,
                record_class=self._rg_dataset_type._RECORD_TYPE,
                dataset=self.dataset_full_prepared,
                multi_label=self._multi_label,
                settings=self._settings,
                seed=self._seed,
                model=self.model,
            )
        elif framework is Framework.TRANSFORMERS:
            from argilla.training.transformers import ArgillaTransformersTrainer

            self._trainer = ArgillaTransformersTrainer(
                name=self._name,
                workspace=self._workspace,
                record_class=self._rg_dataset_type._RECORD_TYPE,
                dataset=self.dataset_full_prepared,
                multi_label=self._multi_label,
                settings=self._settings,
                seed=self._seed,
                model=self.model,
            )
        elif framework is Framework.PEFT:
            from argilla.training.peft import ArgillaPeftTrainer

            self._trainer = ArgillaPeftTrainer(
                name=self._name,
                workspace=self._workspace,
                record_class=self._rg_dataset_type._RECORD_TYPE,
                dataset=self.dataset_full_prepared,
                multi_label=self._multi_label,
                settings=self._settings,
                seed=self._seed,
                model=self.model,
            )
        elif framework is Framework.SPACY:
            from argilla.training.spacy import ArgillaSpaCyTrainer

            self._trainer = ArgillaSpaCyTrainer(
                name=self._name,
                workspace=self._workspace,
                record_class=self._rg_dataset_type._RECORD_TYPE,
                dataset=self.dataset_full_prepared,
                model=self.model,
                multi_label=self._multi_label,
                settings=self._settings,
                seed=self._seed,
                gpu_id=gpu_id,
                **framework_kwargs,  # freeze_tok2vec
            )
        elif framework is Framework.SPACY_TRANSFORMERS:
            from argilla.training.spacy import ArgillaSpaCyTransformersTrainer

            self._trainer = ArgillaSpaCyTransformersTrainer(
                name=self._name,
                workspace=self._workspace,
                record_class=self._rg_dataset_type._RECORD_TYPE,
                dataset=self.dataset_full_prepared,
                model=self.model,
                multi_label=self._multi_label,
                settings=self._settings,
                seed=self._seed,
                gpu_id=gpu_id,
                **framework_kwargs,  # update_transformer
            )
        elif framework is Framework.OPENAI:
            from argilla.training.openai import ArgillaOpenAITrainer

            self._trainer = ArgillaOpenAITrainer(
                name=self._name,
                workspace=self._workspace,
                record_class=self._rg_dataset_type._RECORD_TYPE,
                dataset=self.dataset_full_prepared,
                model=self.model,
                multi_label=self._multi_label,
                settings=self._settings,
                seed=self._seed,
            )
        elif framework is Framework.SPAN_MARKER:
            if self._rg_dataset_type is not DatasetForTokenClassification:
                raise NotImplementedError(f"{Framework.SPAN_MARKER} only supports `TokenClassification` tasks.")
            from argilla.training.span_marker import ArgillaSpanMarkerTrainer

            self._trainer = ArgillaSpanMarkerTrainer(
                name=self._name,
                workspace=self._workspace,
                record_class=self._rg_dataset_type._RECORD_TYPE,
                dataset=self.dataset_full_prepared,
                model=self.model,
                multi_label=self._multi_label,
                settings=self._settings,
                seed=self._seed,
            )

        self._logger.info(self)
        self._track_trainer_usage(framework=framework, task=self._rg_dataset_type._RECORD_TYPE.__name__)

    def _track_trainer_usage(self, framework: str, task: str):
        self._CLIENT.track_data(action="ArgillaTrainerUsage", data={"framework": framework, "task": task})

    def __repr__(self) -> str:
        """
        `trainer.__repr__()` prints out the trainer's parameters and a summary of how to use the trainer

        Returns:
          The trainer object.
        """
        return f"""\
ArgillaBaseTrainer info:
_________________________________________________________________
These baseline params are fixed:
    dataset: {self._name}
    task: {self._rg_dataset_type.__name__}
    multi_label: {self._multi_label}
    train_size: {self._train_size}
    seed: {self._seed}

{self._trainer.__class__} info:
_________________________________________________________________
The parameters are configurable via `trainer.update_config()`:
    {self._trainer}

Using the trainer:
_________________________________________________________________
`trainer.train(output_dir)` to train to start training. `output_dir` is the directory to save the model automatically.
`trainer.predict(text, as_argilla_records=True)` to make predictions.
`trainer.save(output_dir)` to save the model manually."""

    def update_config(self, *args, **kwargs):
        """
        It updates the configuration of the trainer, but the parameters depend on the trainer.subclass.
        """
        self._trainer.update_config(*args, **kwargs)
        self._logger.info(
            "Updated parameters:\n"
            + "_________________________________________________________________\n"
            + f"{self._trainer}"
        )

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        `predict` takes a string or list of strings and returns a list of dictionaries, each dictionary
        containing the text, the predicted label, and the confidence score.

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the output will be a list of Argilla records instead of dictionaries. Defaults to True.

        Returns:
          A list of predictions or Argilla records.
        """
        return self._trainer.predict(text=text, as_argilla_records=as_argilla_records, **kwargs)

    def train(self, output_dir: str):
        """
        `train` takes in a path to a file and trains the model. If a path is provided,
        the model is saved to that path.

        Args:
          output_dir (str): The path to the model file.
        """
        self._trainer.train(output_dir)

    def save(self, output_dir: str):
        """
        Saves the model to the specified path.

        Args:
          output_dir (str): The path to the directory where the model will be saved.
        """
        self._trainer.save(output_dir)


class ArgillaTrainerSkeleton(ABC):
    def __init__(
        self,
        name: str,
        dataset,
        record_class: Union[TokenClassificationRecord, Text2TextRecord, TextClassificationRecord],
        workspace: Optional[str] = None,
        multi_label: bool = False,
        settings: Union[TextClassificationSettings, TokenClassificationSettings] = None,
        model: str = None,
        seed: int = None,
        *arg,
        **kwargs,
    ):
        self._name = name
        self._workspace = workspace or get_workspace()
        self._dataset = dataset
        self._record_class = record_class
        self._multi_label = multi_label
        self._settings = settings
        if self._settings:
            self._label2id = self._settings.label2id or None
            self._id2label = self._settings.id2label
            self._label_list = list(self._label2id.keys())
        else:
            self._label_list = None
            self._label2id = None
            self._id2label = None
        self._model = model
        self._seed = seed

    @abstractmethod
    def init_training_args(self):
        """
        Initializes the training arguments.
        """

    @abstractmethod
    def init_model(self):
        """
        Initializes a model.
        """

    @abstractmethod
    def update_config(self, *args, **kwargs):
        """
        Updates the configuration of the trainer, but the parameters depend on the trainer.subclass.
        """

    @abstractmethod
    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        Predicts the label of the text.
        """

    @abstractmethod
    def train(self, output_dir: str = None):
        """
        Trains the model.
        """

    @abstractmethod
    def save(self, output_dir: str):
        """
        Saves the model to the specified path.
        """



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime, timedelta
from typing import Optional

from fastapi import APIRouter, Depends
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm, SecurityScopes
from jose import JWTError, jwt
from sqlalchemy.ext.asyncio import AsyncSession

from argilla.server.contexts import accounts
from argilla.server.database import get_async_db
from argilla.server.errors import UnauthorizedError
from argilla.server.models import User
from argilla.server.security.auth_provider.base import AuthProvider, api_key_header
from argilla.server.security.model import Token

from .settings import Settings
from .settings import settings as local_security

_oauth2_scheme = OAuth2PasswordBearer(tokenUrl=local_security.public_oauth_token_url, auto_error=False)


class LocalAuthProvider(AuthProvider):
    def __init__(self, settings: Settings):
        self.router = APIRouter(tags=["security"])
        self.settings = settings

        # TODO: maybe it's better if we move this endpoint to apis/v0/handlers
        @self.router.post(
            settings.token_api_url,
            response_model=Token,
            operation_id="login_for_access_token",
        )
        async def login_for_access_token(
            db: AsyncSession = Depends(get_async_db),
            form_data: OAuth2PasswordRequestForm = Depends(),
        ) -> Token:
            user = await accounts.authenticate_user(db, form_data.username, form_data.password)
            if not user:
                raise UnauthorizedError()
            access_token_expires = timedelta(minutes=self.settings.token_expiration_in_minutes)
            access_token = self._create_access_token(user.username, expires_delta=access_token_expires)
            return Token(access_token=access_token)

    def _create_access_token(self, username: str, expires_delta: Optional[timedelta] = None) -> str:
        """
        Creates an access token

        Parameters
        ----------
        username:
            The user name
        expires_delta:
            Token expiration

        Returns
        -------
            An access token string
        """
        to_encode = {
            "sub": username,
        }
        if expires_delta:
            to_encode["exp"] = datetime.utcnow() + expires_delta

        return jwt.encode(
            to_encode,
            self.settings.secret_key,
            algorithm=self.settings.algorithm,
        )

    async def fetch_token_user(self, db: AsyncSession, token: str) -> Optional[User]:
        """
        Fetch the user for a given access token

        Parameters
        ----------
        token:
            The access token

        Returns
        -------
            An User instance if a valid token was provided. None otherwise
        """
        try:
            payload = jwt.decode(
                token,
                self.settings.secret_key,
                algorithms=[self.settings.algorithm],
            )
            username: str = payload.get("sub")
            if username:
                user = await accounts.get_user_by_username(db, username)
                return user
        except JWTError:
            return None

    async def get_current_user(
        self,
        security_scopes: SecurityScopes,
        db: AsyncSession = Depends(get_async_db),
        api_key: Optional[str] = Depends(api_key_header),
        token: Optional[str] = Depends(_oauth2_scheme),
    ) -> User:
        user = None

        if api_key:
            user = await accounts.get_user_by_api_key(db, api_key)
        elif token:
            user = await self.fetch_token_user(db, token)

        if user is None:
            raise UnauthorizedError()

        return user


def create_local_auth_provider():
    settings = Settings()

    return LocalAuthProvider(settings=settings)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
from typing import Any, Callable, Dict

from pydantic import BaseModel, PrivateAttr


# TODO(@frascuchon): Define as dataclasses.dataclass
class MetricSummary(BaseModel):
    """THe metric summary result data model"""

    data: Dict[str, Any]
    _build_visualization: Callable = PrivateAttr()

    def visualize(self):
        try:
            return self._build_visualization()
        except ModuleNotFoundError:
            warnings.warn("Please, install plotly in order to use this feature:\n> pip install plotly", stacklevel=2)

    @classmethod
    def new_summary(cls, data: Dict[str, Any], visualization: Callable) -> "MetricSummary":
        summary = cls(data=data)
        summary._build_visualization = visualization
        return summary



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Awaitable, Callable
from uuid import UUID

from sqlalchemy.ext.asyncio import async_object_session

from argilla.server.contexts import accounts
from argilla.server.daos.models.datasets import DatasetDB
from argilla.server.errors import ForbiddenOperationError
from argilla.server.models import (
    Dataset,
    Field,
    Question,
    Record,
    Response,
    Suggestion,
    User,
    UserRole,
    Workspace,
    WorkspaceUser,
)

PolicyAction = Callable[[User], Awaitable[bool]]


async def _exists_workspace_user_by_user_and_workspace_id(user: User, workspace_id: UUID) -> bool:
    db = async_object_session(user)
    workspace = await accounts.get_workspace_user_by_workspace_id_and_user_id(db, workspace_id, user.id)
    return workspace is not None


async def _exists_workspace_user_by_user_and_workspace_name(user: User, workspace_name: str) -> bool:
    db = async_object_session(user)
    workspace = await accounts.get_workspace_by_name(db, workspace_name)
    if workspace is None:
        return False
    return await accounts.get_workspace_user_by_workspace_id_and_user_id(db, workspace.id, user.id) is not None


class WorkspaceUserPolicy:
    @classmethod
    def list(cls, workspace_id: UUID) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, workspace_id)
            )

        return is_allowed

    @classmethod
    async def create(cls, actor: User) -> bool:
        return actor.is_owner

    @classmethod
    def delete(cls, workspace_user: WorkspaceUser) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.role == UserRole.owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, workspace_user.workspace_id)
            )

        return is_allowed


class WorkspacePolicy:
    @classmethod
    async def list(cls, actor: User) -> bool:
        return True

    @classmethod
    async def create(cls, actor: User) -> bool:
        return actor.is_owner

    @classmethod
    def delete(cls, workspace: Workspace) -> PolicyAction:
        def is_allowed(actor: User) -> bool:
            return actor.is_owner

        return is_allowed


class WorkspacePolicyV1:
    @classmethod
    def get(cls, workspace_id: UUID) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or await _exists_workspace_user_by_user_and_workspace_id(actor, workspace_id)

        return is_allowed

    @classmethod
    async def delete(cls, actor: User) -> bool:
        return actor.is_owner

    @classmethod
    async def list_workspaces_me(cls, actor: User) -> bool:
        return True


class UserPolicy:
    @classmethod
    async def list(cls, actor: User) -> bool:
        return actor.is_owner

    @classmethod
    async def create(cls, actor: User) -> bool:
        return actor.is_owner

    @classmethod
    def delete(cls, user: User) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner

        return is_allowed


class UserPolicyV1:
    @classmethod
    async def list_workspaces(cls, actor: User) -> bool:
        return actor.is_owner


class DatasetPolicy:
    @classmethod
    async def list(cls, user: User) -> bool:
        return True

    @classmethod
    def get(cls, dataset: DatasetDB) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)

        return is_allowed

    @classmethod
    def create(cls, workspace_name: str) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_name(actor, workspace_name)
            )

        return is_allowed

    @classmethod
    def update(cls, dataset: DatasetDB) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)

        return is_allowed

    @classmethod
    def delete(cls, dataset: DatasetDB) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)
            )

        return is_allowed

    @classmethod
    def open(cls, dataset: DatasetDB) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)
            )

        return is_allowed

    @classmethod
    def close(cls, dataset: DatasetDB) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)
            )

        return is_allowed

    @classmethod
    def copy(cls, dataset: DatasetDB, target_workspace: Workspace) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, target_workspace.id)
                and await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)
            )

        return is_allowed

    @classmethod
    def delete_records(cls, dataset: DatasetDB) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)
            )

        return is_allowed


class DatasetPolicyV1:
    @classmethod
    async def list(cls, actor: User) -> bool:
        return True

    @classmethod
    def get(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)

        return is_allowed

    @classmethod
    def list_dataset_records_with_all_responses(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def create(cls, workspace_id: UUID) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, workspace_id)
            )

        return is_allowed

    @classmethod
    def create_field(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def create_question(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def create_records(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def delete_records(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def search_records(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)

        return is_allowed

    @classmethod
    def publish(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def delete(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def update(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_id(actor, dataset.workspace_id)
            )

        return is_allowed


class FieldPolicyV1:
    @classmethod
    def update(cls, field: Field) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, field.dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def delete(cls, field: Field) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, field.dataset.workspace_id)
            )

        return is_allowed


class QuestionPolicyV1:
    @classmethod
    def update(cls, question: Question) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, question.dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def delete(cls, question: Question) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, question.dataset.workspace_id)
            )

        return is_allowed


class RecordPolicyV1:
    @classmethod
    def delete(cls, record: Record) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, record.dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def create_response(cls, record: Record) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or await _exists_workspace_user_by_user_and_workspace_id(
                actor, record.dataset.workspace_id
            )

        return is_allowed

    @classmethod
    def get_suggestions(cls, record: Record) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or await _exists_workspace_user_by_user_and_workspace_id(
                actor, record.dataset.workspace_id
            )

        return is_allowed

    @classmethod
    def create_suggestion(cls, record: Record) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, record.dataset.workspace_id)
            )

        return is_allowed

    @classmethod
    def delete_suggestions(cls, record: Record) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, record.dataset.workspace_id)
            )

        return is_allowed


class ResponsePolicyV1:
    @classmethod
    def update(cls, response: Response) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return (
                actor.is_owner
                or actor.id == response.user_id
                or (
                    actor.is_admin
                    and await _exists_workspace_user_by_user_and_workspace_id(
                        actor, response.record.dataset.workspace_id
                    )
                )
            )

        return is_allowed

    @classmethod
    def delete(cls, response: Response) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return (
                actor.is_owner
                or actor.id == response.user_id
                or (
                    actor.is_admin
                    and await _exists_workspace_user_by_user_and_workspace_id(
                        actor, response.record.dataset.workspace_id
                    )
                )
            )

        return is_allowed


class SuggestionPolicyV1:
    @classmethod
    def delete(cls, suggestion: Suggestion) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin
                and await _exists_workspace_user_by_user_and_workspace_id(actor, suggestion.record.dataset.workspace_id)
            )

        return is_allowed


class DatasetSettingsPolicy:
    @classmethod
    def list(cls, dataset: DatasetDB) -> PolicyAction:
        return DatasetPolicy.get(dataset)

    @classmethod
    def save(cls, dataset: DatasetDB) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)
            )

        return is_allowed

    @classmethod
    def delete(cls, dataset: Dataset) -> PolicyAction:
        async def is_allowed(actor: User) -> bool:
            return actor.is_owner or (
                actor.is_admin and await _exists_workspace_user_by_user_and_workspace_name(actor, dataset.workspace)
            )

        return is_allowed


async def authorize(actor: User, policy_action: PolicyAction) -> None:
    if not await is_authorized(actor, policy_action):
        raise ForbiddenOperationError()


async def is_authorized(actor: User, policy_action: PolicyAction) -> bool:
    return await policy_action(actor)
